# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""mycube's __init__

"""



# copyright 2003 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.server.migractions"""

import os.path as osp

from cubicweb import ExecutionError
from cubicweb.devtools import startpgcluster, stoppgcluster
from cubicweb.server.sqlutils import SQL_PREFIX
from cubicweb.server.test_migractions.unittest_migractions import MigrationTC

HERE = osp.dirname(osp.abspath(__file__))
migrschema = None


def setUpModule():
    startpgcluster(__file__)


def tearDownModule(*args):
    global migrschema
    del migrschema
    if hasattr(MigrationCommandsComputedTC, "origschema"):
        del MigrationCommandsComputedTC.origschema
    stoppgcluster(__file__)


class MigrationCommandsComputedTC(MigrationTC):
    """Unit tests for computed relations and attributes"""

    appid = "datacomputed"

    def setUp(self):
        MigrationTC.setUp(self)
        # ensure vregistry is reloaded, needed by generated hooks for computed
        # attributes
        self.repo.vreg.set_schema(self.repo.schema)

    def test_computed_relation_add_relation_definition(self):
        self.assertNotIn("works_for", self.schema)
        with self.mh() as (cnx, mh):
            with self.assertRaises(ExecutionError) as exc:
                mh.cmd_add_relation_definition("Employee", "works_for", "Company")
        self.assertEqual(
            str(exc.exception),
            "Cannot add a relation definition for a computed " "relation (works_for)",
        )

    def test_computed_relation_drop_relation_definition(self):
        self.assertIn("notes", self.schema)
        with self.mh() as (cnx, mh):
            with self.assertRaises(ExecutionError) as exc:
                mh.cmd_drop_relation_definition("Company", "notes", "Note")
        self.assertEqual(
            str(exc.exception),
            "Cannot drop a relation definition for a computed " "relation (notes)",
        )

    def test_computed_relation_add_relation_type(self):
        self.assertNotIn("works_for", self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_add_relation_type("works_for")
            self.assertIn("works_for", self.schema)
            self.assertEqual(
                self.schema["works_for"].rule,
                "O employees S, NOT EXISTS (O associates S)",
            )
            self.assertEqual(self.schema["works_for"].objects(), ("Company",))
            self.assertEqual(self.schema["works_for"].subjects(), ("Employee",))
            self.assertFalse(self.table_sql(mh, "works_for_relation"))
            e = cnx.create_entity("Employee")
            a = cnx.create_entity("Employee")
            cnx.create_entity("Company", employees=e, associates=a)
            cnx.commit()
            company = cnx.execute("Company X").get_entity(0, 0)
            self.assertEqual([e.eid], [x.eid for x in company.reverse_works_for])
            mh.rollback()

    def test_computed_relation_drop_relation_type(self):
        self.assertIn("notes", self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_drop_relation_type("notes")
        self.assertNotIn("notes", self.schema)

    def test_computed_relation_sync_schema_props_perms(self):
        self.assertIn("whatever", self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_sync_schema_props_perms("whatever")
            self.assertEqual(
                self.schema["whatever"].rule, "S employees E, O associates E"
            )
            self.assertEqual(self.schema["whatever"].objects(), ("Company",))
            self.assertEqual(self.schema["whatever"].subjects(), ("Company",))
            self.assertFalse(self.table_sql(mh, "whatever_relation"))

    def test_computed_relation_sync_schema_props_perms_security(self):
        with self.mh() as (cnx, mh):
            rdef = next(iter(self.schema["perm_changes"].relation_definitions.values()))
            self.assertEqual(
                rdef.permissions,
                {"add": (), "delete": (), "read": ("managers", "users")},
            )
            mh.cmd_sync_schema_props_perms("perm_changes")
            self.assertEqual(
                self.schema["perm_changes"].permissions, {"read": ("managers",)}
            )
            rdef = next(iter(self.schema["perm_changes"].relation_definitions.values()))
            self.assertEqual(
                rdef.permissions, {"add": (), "delete": (), "read": ("managers",)}
            )

    def test_computed_relation_sync_schema_props_perms_on_rdef(self):
        self.assertIn("whatever", self.schema)
        with self.mh() as (cnx, mh):
            with self.assertRaises(ExecutionError) as exc:
                mh.cmd_sync_schema_props_perms(("Company", "whatever", "Person"))
        self.assertEqual(
            str(exc.exception),
            "Cannot synchronize a relation definition for a computed "
            "relation (whatever)",
        )

    def test_computed_relation_rename_relation_type(self):
        with self.mh() as (cnx, mh):
            mh.cmd_rename_relation_type("to_be_renamed", "renamed")
        self.assertIn("renamed", self.schema)
        self.assertNotIn("to_be_renamed", self.schema)

    # computed attributes migration ############################################

    def setup_add_score(self):
        with self.admin_access.client_cnx() as cnx:
            assert not cnx.execute("Company X")
            c = cnx.create_entity("Company")
            e1 = cnx.create_entity("Employee", reverse_employees=c)
            cnx.create_entity("Note", note=2, concerns=e1)
            e2 = cnx.create_entity("Employee", reverse_employees=c)
            cnx.create_entity("Note", note=4, concerns=e2)
            cnx.commit()

    def assert_score_initialized(self, mh):
        self.assertEqual(
            self.schema["score"].relation_definitions["Company", "Float"].formula,
            "Any AVG(NN) WHERE X employees E, N concerns E, N note NN",
        )
        fields = self.table_schema(mh, f"{SQL_PREFIX}Company")
        self.assertEqual(fields[f"{SQL_PREFIX}score"], ("double precision", None))
        self.assertEqual(
            [[3.0]], mh.rqlexec("Any CS WHERE C score CS, C is Company").rows
        )

    def test_computed_attribute_add_relation_type(self):
        self.assertNotIn("score", self.schema)
        self.setup_add_score()
        with self.mh() as (cnx, mh):
            mh.cmd_add_relation_type("score")
            self.assertIn("score", self.schema)
            self.assertEqual(self.schema["score"].objects(), ("Float",))
            self.assertEqual(self.schema["score"].subjects(), ("Company",))
            self.assert_score_initialized(mh)

    def test_computed_attribute_add_attribute(self):
        self.assertNotIn("score", self.schema)
        self.setup_add_score()
        with self.mh() as (cnx, mh):
            mh.cmd_add_attribute("Company", "score")
            self.assertIn("score", self.schema)
            self.assert_score_initialized(mh)

    def assert_computed_attribute_dropped(self):
        self.assertNotIn("note20", self.schema)
        with self.mh() as (cnx, mh):
            fields = self.table_schema(mh, f"{SQL_PREFIX}Note")
        self.assertNotIn(f"{SQL_PREFIX}note20", fields)

    def test_computed_attribute_drop_type(self):
        self.assertIn("note20", self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_drop_relation_type("note20")
        self.assert_computed_attribute_dropped()

    def test_computed_attribute_drop_relation_definition(self):
        self.assertIn("note20", self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_drop_relation_definition("Note", "note20", "Int")
        self.assert_computed_attribute_dropped()

    def test_computed_attribute_drop_attribute(self):
        self.assertIn("note20", self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_drop_attribute("Note", "note20")
        self.assert_computed_attribute_dropped()

    def test_computed_attribute_sync_schema_props_perms_rtype(self):
        self.assertIn("note100", self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_sync_schema_props_perms("note100")
        rdef = self.schema["note100"].relation_definitions["Note", "Int"]
        self.assertEqual(
            rdef.formula_select.as_string(), "Any (N * 100) WHERE X note N, X is Note"
        )
        self.assertEqual(rdef.formula, "Any N*100 WHERE X note N")

    def test_computed_attribute_sync_schema_props_perms_rdef(self):
        self.setup_add_score()
        with self.mh() as (cnx, mh):
            mh.cmd_sync_schema_props_perms(("Note", "note100", "Int"))
            self.assertEqual(
                [[200], [400]], cnx.execute("Any N ORDERBY N WHERE X note100 N").rows
            )
            self.assertEqual(
                [[300]], cnx.execute("Any CS WHERE C score100 CS, C is Company").rows
            )


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Fake objects to ease testing of cubicweb without a fully working environment
"""

from contextlib import contextmanager

from logilab.database import get_db_helper

from cubicweb.cwvreg import CWRegistryStore
from cubicweb.devtools import BASE_URL
from cubicweb.devtools.apptest_config import ApptestConfiguration
from cubicweb.req import RequestSessionAndConnectionBase


class FakeConfig(dict, ApptestConfiguration):
    translations = {}
    uiprops = {}
    apphome = None
    debugmode = False

    def __init__(self, appid="data", apphome=None, cubes=()):
        self.appid = appid
        self.apphome = apphome
        self._cubes = cubes
        self["auth-mode"] = "cookie"
        self["uid"] = None
        self["base-url"] = BASE_URL
        self["rql-cache-size"] = 3000
        self.datadir_url = BASE_URL + "data/"

    def cubes(self, expand=False):
        return self._cubes

    def sources(self):
        return {"system": {"db-driver": "sqlite"}}


class FakeCWRegistryStore(CWRegistryStore):
    def property_value(self, key):
        if key == "ui.language":
            return "en"
        assert False


class FakeUser:
    login = "toto"
    eid = 0

    def in_groups(self, groups):
        return True


class FakeConnection(RequestSessionAndConnectionBase):
    def __init__(self, repo=None, user=None, vreg=None):
        self.repo = repo
        if vreg is None:
            vreg = getattr(self.repo, "vreg", None)
        if vreg is None:
            vreg = FakeCWRegistryStore(FakeConfig(), initlog=False)
        self.vreg = vreg
        self.cnxset = FakeConnectionsSet()
        self.user = user or FakeUser()
        self.is_internal_session = False
        self.transaction_data = {}

    def execute(self, *args, **kwargs):
        pass

    def commit(self, *args):
        self.transaction_data.clear()

    def system_sql(self, sql, args=None):
        pass

    def set_entity_cache(self, entity):
        pass

    def security_enabled(self, read=False, write=False):
        class FakeCM:
            def __enter__(self):
                pass

            def __exit__(self, exctype, exc, traceback):
                pass

        return FakeCM()

    # for use with enabled_security context manager
    read_security = write_security = True

    @contextmanager
    def running_hooks_ops(self):
        yield


class FakeRepo:
    querier = None

    def __init__(self, schema, vreg=None, config=None):
        self.eids = {}
        self._count = 0
        self.schema = schema
        self.config = config or FakeConfig()
        self.vreg = vreg or FakeCWRegistryStore(self.config, initlog=False)
        self.vreg.schema = schema


class FakeSource:
    dbhelper = get_db_helper("sqlite")

    def __init__(self, uri):
        self.uri = uri


class FakeConnectionsSet:
    def source(self, uri):
        return FakeSource(uri)



# copyright 2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""A rql query completer. Given the start of a RQL query, provides possible
completions.
"""
from logging import getLogger

from rql import parse
from rql.utils import rqlvar_maker
from yams.interfaces import IVocabularyConstraint


class RQLSuggestionsBuilder:
    """main entry point is `build_suggestions()` which takes
    an incomplete RQL query and returns a list of suggestions to complete
    the query.

    .. automethod:: cubicweb.rqlsuggestions.RQLSuggestionsBuilder.build_suggestions
    """

    # maximum number of results to fetch when suggesting attribute values
    attr_value_limit = 20

    def __init__(self, req):
        self.req = req
        self.logger = getLogger("cubicweb.rqlsuggestions")

    def build_suggestions(self, user_rql):
        """return a list of suggestions to complete `user_rql`

        :param user_rql: an incomplete RQL query
        """
        req = self.req
        try:
            if (
                "WHERE" not in user_rql
            ):  # don't try to complete if there's no restriction
                return []
            variables, restrictions = [
                part.strip() for part in user_rql.split("WHERE", 1)
            ]
            if "," in restrictions:
                restrictions, incomplete_part = restrictions.rsplit(",", 1)
                user_rql = f"{variables} WHERE {restrictions}"
            else:
                restrictions, incomplete_part = "", restrictions
                user_rql = variables
            select = parse(user_rql, print_errors=False).children[0]
            req.vreg.rqlhelper.annotate(select)
            req.vreg.compute_var_types(req, select, {})
            if restrictions:
                return [
                    f"{user_rql}, {suggestion}"
                    for suggestion in self._rql_build_suggestions(
                        select, incomplete_part
                    )
                ]
            else:
                return [
                    f"{user_rql} WHERE {suggestion}"
                    for suggestion in self._rql_build_suggestions(
                        select, incomplete_part
                    )
                ]
        except Exception as exc:  # we never want to crash
            print(exc)
            self.logger.debug("failed to build suggestions: %s", exc)
            return []

    # actual completion entry points #########################################
    def _rql_build_suggestions(self, select, incomplete_part):
        """
        :param select: the annotated select node (rql syntax tree)
        :param incomplete_part: the part of the rql query that needs
                                to be completed, (e.g. ``X is Pr``, ``X re``)
        """
        chunks = incomplete_part.split(None, 2)
        if not chunks:  # nothing to complete
            return []
        if len(chunks) == 1:  # `incomplete` looks like "MYVAR"
            return self._complete_rqlvar(select, *chunks)
        elif len(chunks) == 2:  # `incomplete` looks like "MYVAR some_rel"
            return self._complete_rqlvar_and_rtype(select, *chunks)
        elif len(chunks) == 3:  # `incomplete` looks like "MYVAR some_rel something"
            return self._complete_relation_object(select, *chunks)
        else:  # would be anything else, hard to decide what to do here
            return []

    def _complete_rqlvar(self, select, rql_var):
        """return suggestions for "variable only" incomplete_part

        as in :

        - Any X WHERE X
        - Any X WHERE X is Project, Y
        - etc.
        """
        return [
            f"{rql_var} {rtype} {dest_var}"
            for rtype, dest_var in self._possible_relations(select, rql_var)
        ]

    def _complete_rqlvar_and_rtype(self, select, rql_var, user_rtype):
        """return suggestions for "variable + rtype" incomplete_part

        as in :

        - Any X WHERE X is
        - Any X WHERE X is Person, X firstn
        - etc.
        """
        # special case `user_type` == 'is', return every possible type.
        if user_rtype == "is":
            return self._complete_is_relation(select, rql_var)
        else:
            return [
                f"{rql_var} {rtype} {dest_var}"
                for rtype, dest_var in self._possible_relations(select, rql_var)
                if rtype.startswith(user_rtype)
            ]

    def _complete_relation_object(self, select, rql_var, user_rtype, user_value):
        """return suggestions for "variable + rtype + some_incomplete_value"

        as in :

        - Any X WHERE X is Per
        - Any X WHERE X is Person, X firstname "
        - Any X WHERE X is Person, X firstname "Pa
        - etc.
        """
        # special case `user_type` == 'is', return every possible type.
        if user_rtype == "is":
            return self._complete_is_relation(select, rql_var, user_value)
        elif user_value:
            if user_value[0] in ('"', "'"):
                # if finished string, don't suggest anything
                if len(user_value) > 1 and user_value[-1] == user_value[0]:
                    return []
                user_value = user_value[1:]
                return [
                    f'{rql_var} {user_rtype} "{value}"'
                    for value in self._vocabulary(
                        select, rql_var, user_rtype, user_value
                    )
                ]
        return []

    def _complete_is_relation(self, select, rql_var, prefix=""):
        """return every possible types for rql_var

        :param prefix: if specified, will only return entity types starting
                       with the specified value.
        """
        return [
            f"{rql_var} is {etype}"
            for etype in self._possible_etypes(select, rql_var, prefix)
        ]

    def _etypes_suggestion_set(self):
        """return the list of possible entity types to suggest

        The default is to return any non-final entity type available
        in the schema.

        Can be overridden for instance if an application decides
        to restrict this list to a meaningful set of business etypes.
        """
        schema = self.req.vreg.schema
        return {eschema.type for eschema in schema.entities() if not eschema.final}

    def _possible_etypes(self, select, rql_var, prefix=""):
        """return all possible etypes for `rql_var`

        The returned list will always be a subset of meth:`etypes_suggestion_set`

        :param select: the annotated select node (rql syntax tree)
        :param rql_var: the variable name for which we want to know possible types
        :param prefix: if specified, will only return etypes starting with it
        """
        available_etypes = self._etypes_suggestion_set()
        possible_etypes = set()
        for sol in select.solutions:
            if rql_var in sol and sol[rql_var] in available_etypes:
                possible_etypes.add(sol[rql_var])
        if not possible_etypes:
            # `Any X WHERE X is Person, Y is`
            # -> won't have a solution, need to give all etypes
            possible_etypes = available_etypes
        return sorted(etype for etype in possible_etypes if etype.startswith(prefix))

    def _possible_relations(self, select, rql_var, include_meta=False):
        """return a list of couple (rtype, dest_var) for each possible
        relations with `rql_var` as subject.

        ``dest_var`` will be picked among availabel variables if types match,
        otherwise a new one will be created.
        """
        schema = self.req.vreg.schema
        relations = set()
        untyped_dest_var = next(rqlvar_maker(defined=select.defined_vars))
        # for each solution
        # 1. find each possible relation
        # 2. for each relation:
        #    2.1. if the relation is meta, skip it
        #    2.2. for each possible destination type, pick up possible
        #         variables for this type or use a new one
        for sol in select.solutions:
            etype = sol[rql_var]
            sol_by_types = {}
            for varname, var_etype in sol.items():
                # don't push subject var to avoid "X relation X" suggestion
                if varname != rql_var:
                    sol_by_types.setdefault(var_etype, []).append(varname)
            for rschema in list(schema[etype].subject_relations.values()):
                if include_meta or not rschema.meta:
                    for dest in rschema.objects(etype):
                        for varname in sol_by_types.get(dest.type, (untyped_dest_var,)):
                            suggestion = (rschema.type, varname)
                            if suggestion not in relations:
                                relations.add(suggestion)
        return sorted(relations)

    def _vocabulary(self, select, rql_var, user_rtype, rtype_incomplete_value):
        """return acceptable vocabulary for `rql_var` + `user_rtype` in `select`

        Vocabulary is either found from schema (Yams) definition or
        directly from database.
        """
        schema = self.req.vreg.schema
        vocab = []
        for sol in select.solutions:
            # for each solution :
            # - If a vocabulary constraint exists on `rql_var+user_rtype`, use it
            #   to define possible values
            # - Otherwise, query the database to fetch available values from
            #   database (limiting results to `self.attr_value_limit`)
            try:
                eschema = schema.entity_schema_for(sol[rql_var])
                rdef = eschema.relation_definition(user_rtype)
            except KeyError:  # unknown relation
                continue
            cstr = rdef.constraint_by_interface(IVocabularyConstraint)
            if cstr is not None:
                # a vocabulary is found, use it
                vocab += [
                    value
                    for value in cstr.vocabulary()
                    if value.startswith(rtype_incomplete_value)
                ]
            elif rdef.final:
                # no vocab, query database to find possible value
                vocab_rql = "DISTINCT Any V LIMIT {} WHERE X is {}, X {} V".format(
                    self.attr_value_limit,
                    eschema.type,
                    user_rtype,
                )
                vocab_kwargs = {}
                if rtype_incomplete_value:
                    vocab_rql += ", X %s LIKE %%(value)s" % user_rtype
                    vocab_kwargs["value"] = f"{rtype_incomplete_value}%"
                vocab += [value for value, in self.req.execute(vocab_rql, vocab_kwargs)]
        return sorted(set(vocab))



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""
Generalities
------------

Paraphrasing the `emacs`_ documentation, let us say that hooks are an important
mechanism for customizing an application. A hook is basically a list of
functions to be called on some well-defined occasion (this is called `running
the hook`).

.. _`emacs`: https://www.gnu.org/software/emacs/manual/html_node/emacs/Hooks.html

Hooks
~~~~~

In |cubicweb|, hooks are subclasses of the :class:`~cubicweb.server.hook.Hook`
class. They are selected over a set of pre-defined `events` (and possibly more
conditions, hooks being selectable appobjects like views and components).  They
should implement a :meth:`~cubicweb.server.hook.Hook.__call__` method that will
be called when the hook is triggered.

There are two families of events: data events (before / after any individual
update of an entity / or a relation in the repository) and server events (such
as server startup or shutdown).  In a typical application, most of the hooks are
defined over data events.

Also, some :class:`~cubicweb.server.hook.Operation` may be registered by hooks,
which will be fired when the transaction is commited or rolled back.

The purpose of data event hooks is usually to complement the data model as
defined in the schema, which is static by nature and only provide a restricted
builtin set of dynamic constraints, with dynamic or value driven behaviours.
For instance they can serve the following purposes:

* enforcing constraints that the static schema cannot express (spanning several
  entities/relations, exotic value ranges and cardinalities, etc.)

* implement computed attributes

It is functionally equivalent to a `database trigger`_, except that database
triggers definition languages are not standardized, hence not portable (for
instance, PL/SQL works with Oracle and PostgreSQL but not Sqlite).

.. _`database trigger`: http://en.wikipedia.org/wiki/Database_trigger


.. hint::

   It is a good practice to write unit tests for each hook. See an example in
   :ref:`hook_test`

Operations
~~~~~~~~~~

Operations are subclasses of the :class:`~cubicweb.server.hook.Operation` class
that may be created by hooks and scheduled to happen on `precommit`,
`postcommit` or `rollback` event (i.e. respectivly before/after a commit or
before a rollback of a transaction).

Hooks are being fired immediately on data operations, and it is sometime
necessary to delay the actual work down to a time where we can expect all
information to be there, or when all other hooks have run (though take care
since operations may themselves trigger hooks). Also while the order of
execution of hooks is data dependant (and thus hard to predict), it is possible
to force an order on operations.

So, for such case where you may miss some information that may be set later in
the transaction, you should instantiate an operation in the hook.

Operations may be used to:

* implements a validation check which needs that all relations be already set on
  an entity

* process various side effects associated with a transaction such as filesystem
  udpates, mail notifications, etc.


Events
------

Hooks are mostly defined and used to handle `dataflow`_ operations. It
means as data gets in (entities added, updated, relations set or
unset), specific events are issued and the Hooks matching these events
are called.

You can get the event that triggered a hook by accessing its `event`
attribute.

.. _`dataflow`: http://en.wikipedia.org/wiki/Dataflow


Entity modification related events
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

When called for one of these events, hook will have an `entity` attribute
containing the entity instance.

- `before_add_entity`, `before_update_entity`:

  On those events, you can access the modified attributes of the entity using
  the `entity.cw_edited` dictionary. The values can be modified and the old
  values can be retrieved.

  If you modify the `entity.cw_edited` dictionary in the hook, that is before
  the database operations take place, you will avoid the need to process a whole
  new rql query and the underlying backend query (eg usually sql) will contain
  the modified data. For example:

  .. sourcecode:: python

     self.entity.cw_edited['age'] = 42

  will modify the age before it is written to the backend storage.

  Similarly, removing an attribute from `cw_edited` will cancel its
  modification:

  .. sourcecode:: python

     del self.entity.cw_edited['age']

  On a `before_update_entity` event, you can access the old and new values:

  .. sourcecode:: python

     old, new = entity.cw_edited.oldnewvalue('age')

- `after_add_entity`, `after_update_entity`

  On those events, you can get the list of attributes that were modified using
  the `entity.cw_edited` dictionary, but you can not modify it or get the old
  value of an attribute.

- `before_delete_entity`, `after_delete_entity`

  On those events, the entity has no `cw_edited` dictionary.

.. note:: `self.entity.cw_set(age=42)` will set the `age` attribute to
  42. But to do so, it will generate a rql query that will have to be processed,
  hence may trigger some hooks, etc. This could lead to infinitely looping hooks.

Relation modification related events
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

When called for one of these events, hook will have `eidfrom`, `rtype`, `eidto`
attributes containing respectively the eid of the subject entity, the relation
type and the eid of the object entity.

* `before_add_relation`, `before_delete_relation`

  On those events, you can still get the original relation by issuing a rql query.

* `after_add_relation`, `after_delete_relation`

Specific selectors are shipped for these kinds of events, see in particular
:class:`~cubicweb.server.hook.match_rtype`.

Also note that relations can be added or deleted, but not updated.

Non data events
~~~~~~~~~~~~~~~

Hooks called on server start/maintenance/stop event (e.g.
`server_startup`, `server_maintenance`, `before_server_shutdown`,
`server_shutdown`) have a `repo` attribute, but *their `_cw` attribute
is None*.  The `server_startup` is called on regular startup, while
`server_maintenance` is called on cubicweb-ctl upgrade or shell
commands. `server_shutdown` is called anyway but connections to the
native source is impossible; `before_server_shutdown` handles that.

Hooks called on backup/restore event (eg `server_backup`,
`server_restore`) have a `repo` and a `timestamp` attributes, but
*their `_cw` attribute is None*.


API
---

Hooks control
~~~~~~~~~~~~~

It is sometimes convenient to explicitly enable or disable some hooks. For
instance if you want to disable some integrity checking hook. This can be
controlled more finely through the `category` class attribute, which is a string
giving a category name.  One can then uses the
:meth:`~cubicweb.server.session.Connection.deny_all_hooks_but` and
:meth:`~cubicweb.server.session.Connection.allow_all_hooks_but` context managers to
explicitly enable or disable some categories.

The existing categories are:

* ``security``, security checking hooks

* ``worfklow``, workflow handling hooks

* ``metadata``, hooks setting meta-data on newly created entities

* ``notification``, email notification hooks

* ``integrity``, data integrity checking hooks

* ``activeintegrity``, data integrity consistency hooks, that you should **never**
  want to disable

* ``syncsession``, hooks synchronizing existing sessions

* ``syncschema``, hooks synchronizing instance schema (including the physical database)

* ``email``, email address handling hooks

* ``bookmark``, bookmark entities handling hooks


Nothing precludes one to invent new categories and use existing mechanisms to
filter them in or out.


Hooks specific predicates
~~~~~~~~~~~~~~~~~~~~~~~~~
.. autoclass:: cubicweb.server.hook.match_rtype
.. autoclass:: cubicweb.server.hook.match_rtype_sets


Hooks and operations classes
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
.. autoclass:: cubicweb.server.hook.Hook
.. autoclass:: cubicweb.server.hook.Operation
.. autoclass:: cubicweb.server.hook.LateOperation
.. autoclass:: cubicweb.server.hook.DataOperationMixIn
"""

from itertools import chain
from logging import getLogger
from typing import Union, Tuple

from logilab.common.decorators import classproperty, cached
from logilab.common.logging_ext import set_log_methods
from logilab.common.registry import NotPredicate, OrPredicate, objectify_predicate

from cubicweb import RegistryNotFound, server
from cubicweb.appobject import AppObject
from cubicweb.cwvreg import CWRegistry, CWRegistryStore
from cubicweb.predicates import ExpectedValuePredicate, is_instance

ENTITIES_HOOKS = set(
    (
        "before_add_entity",
        "after_add_entity",
        "before_update_entity",
        "after_update_entity",
        "before_delete_entity",
        "after_delete_entity",
    )
)
RELATIONS_HOOKS = set(
    (
        "before_add_relation",
        "after_add_relation",
        "before_delete_relation",
        "after_delete_relation",
    )
)
SYSTEM_HOOKS = set(
    (
        "server_backup",
        "server_restore",
        "server_startup",
        "server_maintenance",
        "server_shutdown",
        "before_server_shutdown",
    )
)

ALL_HOOKS = ENTITIES_HOOKS | RELATIONS_HOOKS | SYSTEM_HOOKS


def _iter_kwargs(entities, eids_from_to, kwargs):
    if not entities and not eids_from_to:
        yield kwargs
    elif entities:
        for entity in entities:
            kwargs["entity"] = entity
            yield kwargs
    else:
        for subject, object in eids_from_to:
            kwargs.update({"eidfrom": subject, "eidto": object})
            yield kwargs


class HooksRegistry(CWRegistry):
    def register(self, obj, **kwargs):
        obj.check_events()
        super(HooksRegistry, self).register(obj, **kwargs)

    def call_hooks(self, event, cnx=None, **kwargs):
        """call `event` hooks for an entity or a list of entities (passed
        respectively as the `entity` or ``entities`` keyword argument).
        """
        kwargs["event"] = event
        if cnx is None:  # True for events such as server_start
            for hook in sorted(
                self.possible_objects(cnx, **kwargs), key=lambda x: x.order
            ):
                hook()
        else:
            if "entities" in kwargs:
                assert (
                    "entity" not in kwargs
                ), 'can\'t pass "entities" and "entity" arguments simultaneously'
                assert (
                    "eids_from_to" not in kwargs
                ), 'can\'t pass "entities" and "eids_from_to" arguments simultaneously'
                entities = kwargs.pop("entities")
                eids_from_to = []
            elif "eids_from_to" in kwargs:
                entities = []
                eids_from_to = kwargs.pop("eids_from_to")
            else:
                entities = []
                eids_from_to = []
            pruned = self.get_pruned_hooks(cnx, event, entities, eids_from_to, kwargs)

            # by default, hooks are executed with security turned off
            with cnx.security_enabled(read=False):
                for _kwargs in _iter_kwargs(entities, eids_from_to, kwargs):
                    hooks = sorted(
                        self.filtered_possible_objects(pruned, cnx, **_kwargs),
                        key=lambda x: x.order,
                    )
                    debug = server.DEBUG & server.DBG_HOOKS
                    with cnx.security_enabled(write=False):
                        with cnx.running_hooks_ops():
                            for hook in hooks:
                                if debug:
                                    print(event, _kwargs, hook)
                                hook()

    def get_pruned_hooks(self, cnx, event, entities, eids_from_to, kwargs):
        """return a set of hooks that should not be considered by filtered_possible objects

        the idea is to make a first pass over all the hooks in the
        registry and to mark put some of them in a pruned list. The
        pruned hooks are the one which:

        * are disabled at the connection level

        * have a selector containing a :class:`match_rtype` or an
          :class:`is_instance` predicate which does not match the rtype / etype
          of the relations / entities for which we are calling the hooks. This
          works because the repository calls the hooks grouped by rtype or by
          etype when using the entities or eids_to_from keyword arguments

        Only hooks with a simple predicate or an AndPredicate of simple
        predicates are considered for disabling.

        """
        if "entity" in kwargs:
            entities = [kwargs["entity"]]
        if len(entities):
            look_for_selector = is_instance
            etype = entities[0].__regid__
        elif "rtype" in kwargs:
            look_for_selector = match_rtype
            etype = None
        else:  # nothing to prune, how did we get there ???
            return set()
        cache_key = (event, kwargs.get("rtype"), etype)
        pruned = cnx.pruned_hooks_cache.get(cache_key)
        if pruned is not None:
            return pruned
        pruned = set()
        cnx.pruned_hooks_cache[cache_key] = pruned
        if look_for_selector is not None:
            for id, hooks in self.items():
                for hook in hooks:
                    enabled_cat, main_filter = hook.filterable_selectors()
                    if enabled_cat is not None:
                        if not enabled_cat(hook, cnx):
                            pruned.add(hook)
                            continue
                    if main_filter is not None:
                        if isinstance(main_filter, match_rtype) and (
                            main_filter.frometypes is not None
                            or main_filter.toetypes is not None
                        ):
                            continue
                        first_kwargs = next(
                            _iter_kwargs(entities, eids_from_to, kwargs)
                        )
                        if not main_filter(hook, cnx, **first_kwargs):
                            pruned.add(hook)
        return pruned

    def filtered_possible_objects(self, pruned, *args, **kwargs):
        for appobjects in self.values():
            if pruned:
                filtered_objects = [obj for obj in appobjects if obj not in pruned]
                if not filtered_objects:
                    continue
            else:
                filtered_objects = appobjects
            obj = self._select_best(filtered_objects, *args, **kwargs)
            if obj is None:
                continue
            yield obj


class HooksManager:
    def __init__(self, vreg):
        self.vreg = vreg

    def call_hooks(self, event, cnx=None, **kwargs):
        try:
            registry = self.vreg[f"{event}_hooks"]
        except RegistryNotFound:
            return  # no hooks for this event
        registry.call_hooks(event, cnx, **kwargs)


for event in ALL_HOOKS:
    CWRegistryStore.REGISTRY_FACTORY[f"{event}_hooks"] = HooksRegistry


# some hook specific predicates #################################################


@objectify_predicate
def enabled_category(cls, req, **kwargs):
    if req is None:
        return True  # XXX how to deactivate server startup / shutdown event
    return req.is_hook_activated(cls)


@objectify_predicate
def issued_from_user_query(cls, req, **kwargs):
    return 0 if req.hooks_in_progress else 1


class rechain:
    def __init__(self, *iterators):
        self.iterators = iterators

    def __iter__(self):
        return iter(chain(*self.iterators))


class match_rtype(ExpectedValuePredicate):
    """accept if the relation type is found in expected ones. Optional
    named parameters `frometypes` and `toetypes` can be used to restrict
    target subject and/or object entity types of the relation.

    :param *expected: possible relation types
    :param frometypes: candidate entity types as subject of relation
    :param toetypes: candidate entity types as object of relation
    """

    def __init__(self, *expected, **more):
        self.expected = expected
        self.frometypes = more.pop("frometypes", None)
        self.toetypes = more.pop("toetypes", None)
        assert not more, f"unexpected kwargs in match_rtype: {more}"

    def __call__(self, cls, req, *args, **kwargs):
        if kwargs.get("rtype") not in self.expected:
            return 0
        if (
            self.frometypes is not None
            and req.entity_type(kwargs["eidfrom"]) not in self.frometypes
        ):
            return 0
        if (
            self.toetypes is not None
            and req.entity_type(kwargs["eidto"]) not in self.toetypes
        ):
            return 0
        return 1


class match_rtype_sets(ExpectedValuePredicate):
    """accept if the relation type is in one of the sets given as initializer
    argument. The goal of this predicate is that it keeps reference to original sets,
    so modification to thoses sets are considered by the predicate. For instance

    .. sourcecode:: python

      MYSET = set()

      class Hook1(Hook):
          __regid__ = 'hook1'
          __select__ = Hook.__select__ & match_rtype_sets(MYSET)
          ...

      class Hook2(Hook):
          __regid__ = 'hook2'
          __select__ = Hook.__select__ & match_rtype_sets(MYSET)

    Client code can now change `MYSET`, this will changes the selection criteria
    of :class:`Hook1` and :class:`Hook1`.
    """

    def __init__(self, *expected):
        self.expected = expected

    def __call__(self, cls, req, *args, **kwargs):
        for rel_set in self.expected:
            if kwargs.get("rtype") in rel_set:
                return 1
        return 0


# base class for hook ##########################################################


class Hook(AppObject):
    """Base class for hook.

    Hooks being appobjects like views, they have a `__regid__` and a `__select__`
    class attribute. Like all appobjects, hooks have the `self._cw` attribute which
    represents the current connection. In entity hooks, a `self.entity` attribute is
    also present.

    The `events` tuple is used by the base class selector to dispatch the hook
    on the right events. It is possible to dispatch on multiple events at once
    if needed (though take care as hook attribute may vary as described above).

    .. Note::

      Do not forget to extend the base class selectors as in:

      .. sourcecode:: python

          class MyHook(Hook):
            __regid__ = 'whatever'
            __select__ = Hook.__select__ & is_instance('Person')

      else your hooks will be called madly, whatever the event.
    """

    __select__ = enabled_category()
    # set this in derivated classes
    events: Union[None, Tuple[str], Tuple[str, str]] = None
    category: Union[None, str] = None
    order = 0
    # stop pylint from complaining about missing attributes in Hooks classes
    eidfrom = eidto = entity = rtype = repo = None

    @classmethod
    @cached
    def filterable_selectors(cls):
        search = cls.__select__.search_selector
        if search((NotPredicate, OrPredicate)):
            return None, None
        enabled_cat = search(enabled_category)
        main_filter = search((is_instance, match_rtype))
        return enabled_cat, main_filter

    @classmethod
    def check_events(cls):
        try:
            for event in cls.events:
                if event not in ALL_HOOKS:
                    raise Exception(
                        f"bad event {event} on {cls.__module__}.{cls.__name__}"
                    )
        except AttributeError:
            raise
        except TypeError:
            raise Exception(
                f"bad .events attribute {cls.events} on {cls.__module__}.{cls.__name__}"
            )

    @classmethod
    def __registered__(cls, reg):
        cls.check_events()

    @classproperty
    def __registries__(cls):
        if cls.events is None:
            return []
        return [f"{ev}_hooks" for ev in cls.events]

    known_args = set(("entity", "rtype", "eidfrom", "eidto", "repo", "timestamp"))

    def __init__(self, req, event, **kwargs):
        for arg in self.known_args:
            if arg in kwargs:
                setattr(self, arg, kwargs.pop(arg))
        super(Hook, self).__init__(req, **kwargs)
        self.event = event


set_log_methods(Hook, getLogger("cubicweb.hook"))


# abtract hooks for relation propagation #######################################
# See example usage in hooks of the nosylist cube


class PropagateRelationHook(Hook):
    """propagate some `main_rtype` relation on entities linked as object of
    `subject_relations` or as subject of `object_relations` (the watched
    relations).

    This hook ensure that when one of the watched relation is added, the
    `main_rtype` relation is added to the target entity of the relation.
    Notice there are no default behaviour defined when a watched relation is
    deleted, you'll have to handle this by yourself.

    You usually want to use the :class:`match_rtype_sets` predicate on concrete
    classes.
    """

    events = ("after_add_relation",)

    # to set in concrete class
    main_rtype = None
    subject_relations = None
    object_relations = None

    def __call__(self):
        assert self.main_rtype
        for eid in (self.eidfrom, self.eidto):
            etype = self._cw.entity_type(eid)
            if (
                self.main_rtype
                not in self._cw.vreg.schema.entity_schema_for(etype).subject_relations
            ):
                return
        if self.rtype in self.subject_relations:
            meid, seid = self.eidfrom, self.eidto
        else:
            assert self.rtype in self.object_relations
            meid, seid = self.eidto, self.eidfrom
        self._cw.execute(
            "SET E %s P WHERE X %s P, X eid %%(x)s, E eid %%(e)s, NOT E %s P"
            % (self.main_rtype, self.main_rtype, self.main_rtype),
            {"x": meid, "e": seid},
        )


class PropagateRelationAddHook(Hook):
    """Propagate to entities at the end of watched relations when a `main_rtype`
    relation is added.

    `subject_relations` and `object_relations` attributes should be specified on
    subclasses and are usually shared references with attributes of the same
    name on :class:`PropagateRelationHook`.

    Because of those shared references, you can use `skip_subject_relations` and
    `skip_object_relations` attributes when you don't want to propagate to
    entities linked through some particular relations.
    """

    events = ("after_add_relation",)

    # to set in concrete class (mandatory)
    subject_relations = None
    object_relations = None
    # to set in concrete class (optionally)
    skip_subject_relations = ()
    skip_object_relations = ()

    def __call__(self):
        eschema = self._cw.vreg.schema.entity_schema_for(
            self._cw.entity_type(self.eidfrom)
        )
        execute = self._cw.execute
        for rel in self.subject_relations:
            if (
                rel in eschema.subject_relations
                and rel not in self.skip_subject_relations
            ):
                execute(
                    "SET R %s P WHERE X eid %%(x)s, P eid %%(p)s, "
                    "X %s R, NOT R %s P" % (self.rtype, rel, self.rtype),
                    {"x": self.eidfrom, "p": self.eidto},
                )
        for rel in self.object_relations:
            if (
                rel in eschema.object_relations
                and rel not in self.skip_object_relations
            ):
                execute(
                    "SET R %s P WHERE X eid %%(x)s, P eid %%(p)s, "
                    "R %s X, NOT R %s P" % (self.rtype, rel, self.rtype),
                    {"x": self.eidfrom, "p": self.eidto},
                )


class PropagateRelationDelHook(PropagateRelationAddHook):
    """Propagate to entities at the end of watched relations when a `main_rtype`
    relation is deleted.

    This is the opposite of the :class:`PropagateRelationAddHook`, see its
    documentation for how to use this class.
    """

    events = ("after_delete_relation",)

    def __call__(self):
        eschema = self._cw.vreg.schema.entity_schema_for(
            self._cw.entity_type(self.eidfrom)
        )
        execute = self._cw.execute
        for rel in self.subject_relations:
            if (
                rel in eschema.subject_relations
                and rel not in self.skip_subject_relations
            ):
                execute(
                    "DELETE R %s P WHERE X eid %%(x)s, P eid %%(p)s, "
                    "X %s R" % (self.rtype, rel),
                    {"x": self.eidfrom, "p": self.eidto},
                )
        for rel in self.object_relations:
            if (
                rel in eschema.object_relations
                and rel not in self.skip_object_relations
            ):
                execute(
                    "DELETE R %s P WHERE X eid %%(x)s, P eid %%(p)s, "
                    "R %s X" % (self.rtype, rel),
                    {"x": self.eidfrom, "p": self.eidto},
                )


# abstract classes for operation ###############################################


class Operation:
    """Base class for operations.

    Operation may be instantiated in the hooks' `__call__` method. It always
    takes a connection object as first argument (accessible as `.cnx` from the
    operation instance), and optionally all keyword arguments needed by the
    operation. These keyword arguments will be accessible as attributes from the
    operation instance.

    An operation is triggered on connections set events related to commit /
    rollback transations. Possible events are:

    * `precommit`:

      the transaction is being prepared for commit. You can freely do any heavy
      computation, raise an exception if the commit can't go. or even add some
      new operations during this phase. If you do anything which has to be
      reverted if the commit fails afterwards (eg altering the file system for
      instance), you'll have to support the 'revertprecommit' event to revert
      things by yourself

    * `revertprecommit`:

      if an operation failed while being pre-commited, this event is triggered
      for all operations which had their 'precommit' event already fired to let
      them revert things (including the operation which made the commit fail)

    * `rollback`:

      the transaction has been either rolled back either:

       * intentionally
       * a 'precommit' event failed, in which case all operations are rolled back
         once 'revertprecommit'' has been called

    * `postcommit`:

      the transaction is over. All the ORM entities accessed by the earlier
      transaction are invalid. If you need to work on the database, you need to
      start a new transaction, for instance using a new internal connection,
      which you will need to commit.

    For an operation to support an event, one has to implement the `<event
    name>_event` method with no arguments.

    The order of operations may be important, and is controlled according to
    the insert_index's method output (whose implementation vary according to the
    base hook class used).
    """

    def __init__(self, cnx, **kwargs):
        self.cnx = cnx
        self.__dict__.update(kwargs)
        self.register(cnx)
        # execution information
        self.processed = None  # 'precommit', 'commit'
        self.failed = False

    def register(self, cnx):
        cnx.add_operation(self, self.insert_index())

    def insert_index(self):
        """return the index of the latest instance which is not a
        LateOperation instance
        """
        # faster by inspecting operation in reverse order for heavy transactions
        i = None
        for i, op in enumerate(reversed(self.cnx.pending_operations)):
            if isinstance(op, (LateOperation, SingleLastOperation)):
                continue
            return -i or None
        if i is None:
            return None
        return -(i + 1)

    def handle_event(self, event):
        """delegate event handling to the operation"""
        getattr(self, event)()

    def precommit_event(self):
        """the observed connections set is preparing a commit"""

    def revertprecommit_event(self):
        """an error went when pre-commiting this operation or a later one

        should revert pre-commit's changes but take care, they may have not
        been all considered if it's this operation which failed
        """

    def rollback_event(self):
        """the observed connections set has been rolled back

        do nothing by default
        """

    def postcommit_event(self):
        """the observed connections set has committed"""

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


set_log_methods(Operation, getLogger("cubicweb.session"))


def _container_add(container, value):
    {set: set.add, list: list.append}[container.__class__](container, value)


class DataOperationMixIn:
    """Mix-in class to ease applying a single operation on a set of data,
    avoiding creating as many operations as there are individual modifications.
    The body of the operation must then iterate over the values that have been
    stored in a single operation instance.

    You should try to use this instead of creating on operation for each
    `value`, since handling operations becomes costly on massive data import.

    Usage looks like:

    .. sourcecode:: python

        class MyEntityHook(Hook):
            __regid__ = 'my.entity.hook'
            __select__ = Hook.__select__ & is_instance('MyEntity')
            events = ('after_add_entity',)

            def __call__(self):
                MyOperation.get_instance(self._cw).add_data(self.entity)


        class MyOperation(DataOperationMixIn, Operation):
            def precommit_event(self):
                for bucket in self.get_data():
                    process(bucket)

    You can modify the `containercls` class attribute, which defines the
    container class that should be instantiated to hold payloads. An instance is
    created on instantiation, and then the :meth:`add_data` method will add the
    given data to the existing container. Default to a `set`. Give `list` if you
    want to keep arrival ordering. You can also use another kind of container
    by redefining :meth:`_build_container` and :meth:`add_data`

    More optional parameters can be given to the `get_instance` operation, that
    will be given to the operation constructor (for obvious reasons those
    parameters should not vary accross different calls to this method for a
    given operation).

    .. Note::
        For sanity reason `get_data` will reset the operation, so that once
        the operation has started its treatment, if some hook want to push
        additional data to this same operation, a new instance will be created
        (else that data has a great chance to be never treated). This implies:

        * you should **always** call `get_data` when starting treatment

        * you should **never** call `get_data` for another reason.
    """

    containercls = set

    @classproperty
    def data_key(cls):
        return ("cw.dataops", cls.__name__)

    @classmethod
    def get_instance(cls, cnx, **kwargs):
        # no need to lock: transaction_data already comes from thread's local storage
        try:
            return cnx.transaction_data[cls.data_key]
        except KeyError:
            op = cnx.transaction_data[cls.data_key] = cls(cnx, **kwargs)
            return op

    def __init__(self, *args, **kwargs):
        super(DataOperationMixIn, self).__init__(*args, **kwargs)
        self._container = self._build_container()
        self._processed = False

    def __contains__(self, value):
        return value in self._container

    def _build_container(self):
        return self.containercls()

    def union(self, data):
        """only when container is a set"""
        assert not self._processed, """Trying to add data to a closed operation.
Iterating over operation data closed it and should be reserved to precommit /
postcommit method of the operation."""
        self._container |= data

    def add_data(self, data):
        assert not self._processed, """Trying to add data to a closed operation.
Iterating over operation data closed it and should be reserved to precommit /
postcommit method of the operation."""
        _container_add(self._container, data)

    def remove_data(self, data):
        assert not self._processed, """Trying to add data to a closed operation.
Iterating over operation data closed it and should be reserved to precommit /
postcommit method of the operation."""
        self._container.remove(data)

    def get_data(self):
        assert not self._processed, """Trying to get data from a closed operation.
Iterating over operation data closed it and should be reserved to precommit /
postcommit method of the operation."""
        self._processed = True
        op = self.cnx.transaction_data.pop(self.data_key)
        assert (
            op is self
        ), "Bad handling of operation data, found %s instead of %s for key %s" % (
            op,
            self,
            self.data_key,
        )
        return self._container


class LateOperation(Operation):
    """special operation which should be called after all possible (ie non late)
    operations
    """

    def insert_index(self):
        """return the index of  the lastest instance which is not a
        SingleLastOperation instance
        """
        # faster by inspecting operation in reverse order for heavy transactions
        i = None
        for i, op in enumerate(reversed(self.cnx.pending_operations)):
            if isinstance(op, SingleLastOperation):
                continue
            return -i or None
        if i is None:
            return None
        return -(i + 1)


class SingleLastOperation(Operation):
    """special operation which should be called once and after all other
    operations
    """

    def register(self, cnx):
        """override register to handle cases where this operation has already
        been added
        """
        operations = cnx.pending_operations
        index = self.equivalent_index(operations)
        if index is not None:
            equivalent = operations.pop(index)
        else:
            equivalent = None
        cnx.add_operation(self, self.insert_index())
        return equivalent

    def equivalent_index(self, operations):
        """return the index of the equivalent operation if any"""
        for i, op in enumerate(reversed(operations)):
            if op.__class__ is self.__class__:
                return -(i + 1)
        return None

    def insert_index(self):
        return None


class SendMailOp(SingleLastOperation):
    def __init__(self, cnx, msg=None, recipients=None, **kwargs):
        # may not specify msg yet, as
        # `cubicweb.sobjects.supervision.SupervisionMailOp`
        if msg is not None:
            assert recipients
            self.to_send = [(msg, recipients)]
        else:
            assert recipients is None
            self.to_send = []
        super(SendMailOp, self).__init__(cnx, **kwargs)

    def register(self, cnx):
        previous = super(SendMailOp, self).register(cnx)
        if previous:
            self.to_send = previous.to_send + self.to_send

    def postcommit_event(self):
        self.cnx.repo.threaded_task(self.sendmails)

    def sendmails(self):
        self.cnx.vreg.config.sendmails(self.to_send)


class RQLPrecommitOperation(Operation):
    # to be defined in concrete classes
    rqls = None

    def precommit_event(self):
        execute = self.cnx.execute
        for rql in self.rqls:
            execute(*rql)


class CleanupNewEidsCacheOp(DataOperationMixIn, SingleLastOperation):
    """on rollback of a insert query we have to remove from repository's
    type/source cache eids of entities added in that transaction.

    NOTE: querier's rqlst/solutions cache may have been polluted too with
    queries such as Any X WHERE X eid 32 if 32 has been rolled back however
    generated queries are unpredictable and analysing all the cache probably
    too expensive. Notice that there is no pb when using args to specify eids
    instead of giving them into the rql string.
    """

    data_key = "neweids"

    def rollback_event(self):
        """the observed connections set has been rolled back,
        remove inserted eid from repository type/source cache
        """
        try:
            self.cnx.repo.clear_caches(self.get_data())
        except KeyError:
            pass


class CleanupDeletedEidsCacheOp(DataOperationMixIn, SingleLastOperation):
    """on commit of delete query, we have to remove from repository's
    type/source cache eids of entities deleted in that transaction.
    """

    data_key = "pendingeids"

    def postcommit_event(self):
        """the observed connections set has been rolled back,
        remove inserted eid from repository type/source cache
        """
        try:
            eids = self.get_data()
            self.cnx.repo.clear_caches(eids)
            self.cnx.repo.app_instances_bus.publish(
                ["delete"] + list(str(eid) for eid in eids)
            )
        except KeyError:
            pass



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""helper classes to handle server-side edition of entities"""


from copy import copy

from yams import ValidationError

_MARKER = object()


class dict_protocol_catcher:
    def __init__(self, entity):
        self.__entity = entity

    def __contains__(self, attr):
        return attr in self.__entity.cw_edited

    def __getitem__(self, attr):
        return self.__entity.cw_edited[attr]

    def __setitem__(self, attr, value):
        self.__entity.cw_edited[attr] = value

    def __getattr__(self, attr):
        return getattr(self.__entity, attr)


class EditedEntity(dict):
    """encapsulate entities attributes being written by an RQL query"""

    def __init__(self, entity, **kwargs):
        super().__init__(**kwargs)
        self.entity = entity
        self.skip_security = set()
        self.querier_pending_relations = {}
        self.saved = False

    def __hash__(self):
        # dict|set keyable
        return hash(id(self))

    def __lt__(self, other):
        # we don't want comparison by value inherited from dict
        raise NotImplementedError

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return not (self == other)

    def __setitem__(self, attr, value):
        assert attr != "eid"
        # don't add attribute into skip_security if already in edited
        # attributes, else we may accidentally skip a desired security check
        if attr not in self:
            self.skip_security.add(attr)
        self.edited_attribute(attr, value)

    def __delitem__(self, attr):
        assert not self.saved, "too late to modify edited attributes"
        super().__delitem__(attr)
        self.entity.cw_attr_cache.pop(attr, None)

    def __copy__(self):
        # default copy protocol fails in EditedEntity.__setitem__ because
        # copied entity has no skip_security attribute at this point
        return EditedEntity(self.entity, **self)

    def pop(self, attr, *args):
        # don't update skip_security by design (think to storage api)
        assert not self.saved, "too late to modify edited attributes"
        value = super().pop(attr, *args)
        self.entity.cw_attr_cache.pop(attr, *args)
        return value

    def setdefault(self, attr, default):
        assert attr != "eid"
        # don't add attribute into skip_security if already in edited
        # attributes, else we may accidentally skip a desired security check
        if attr not in self:
            self[attr] = default
        return self[attr]

    def update(self, values, skipsec=True):
        if skipsec:
            setitem = self.__setitem__
        else:
            setitem = self.edited_attribute
        for attr, value in values.items():
            setitem(attr, value)

    def edited_attribute(self, attr, value):
        """attribute being edited by a rql query: should'nt be added to
        skip_security
        """
        assert not self.saved, "too late to modify edited attributes"
        super().__setitem__(attr, value)
        self.entity.cw_attr_cache[attr] = value
        if self.entity._cw.vreg.schema.relation_schema_for(attr).final:
            self.entity._cw_dont_cache_attribute(attr)

    def oldnewvalue(self, attr):
        """returns the couple (old attr value, new attr value)

        NOTE: will only work in a before_update_entity hook
        """
        assert not self.saved, "too late to get the old value"
        # get new value and remove from local dict to force a db query to
        # fetch old value
        newvalue = self.entity.cw_attr_cache.pop(attr, _MARKER)
        oldvalue = getattr(self.entity, attr)
        if newvalue is not _MARKER:
            self.entity.cw_attr_cache[attr] = newvalue
        else:
            newvalue = oldvalue
        return oldvalue, newvalue

    def set_defaults(self):
        """set default values according to the schema"""
        for attr, value in self.entity.e_schema.defaults():
            if attr not in self:
                self[str(attr)] = value

    def check(self, creation=False):
        """check the entity edition against its schema. Only final relation
        are checked here, constraint on actual relations are checked in hooks
        """
        entity = self.entity
        if creation:
            # on creations, we want to check all relations, especially
            # required attributes
            relations = [
                rschema
                for rschema in list(entity.e_schema.subject_relations.values())
                if rschema.final and rschema.type != "eid"
            ]
        else:
            relations = [
                entity._cw.vreg.schema.relation_schema_for(rtype) for rtype in self
            ]
        try:
            entity.e_schema.check(
                dict_protocol_catcher(entity), creation=creation, relations=relations
            )
        except ValidationError as ex:
            ex.entity = self.entity.eid
            raise

    def clone(self):
        thecopy = EditedEntity(copy(self.entity))
        thecopy.entity.cw_attr_cache = copy(self.entity.cw_attr_cache)
        thecopy.entity._cw_related_cache = {}
        thecopy.entity._cw_adapters_cache = {}
        thecopy.update(self, skipsec=False)
        return thecopy



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""This modules defines func / methods for creating test repositories"""

import inspect
import itertools
import logging
from copy import deepcopy
from datetime import datetime, date, time, timedelta, timezone
from decimal import Decimal
from random import randint, choice, sample

from logilab.common import attrdict
from logilab.mtconverter import xml_escape
from rql.utils import decompose_b26 as base_decompose_b26
from yams.constraints import (
    SizeConstraint,
    StaticVocabularyConstraint,
    IntervalBoundConstraint,
    BoundaryConstraint,
    Attribute,
    actual_value,
)

from cubicweb import Binary
from cubicweb.schema import RQLConstraint


def custom_range(start, stop, step):
    if start == stop:
        yield start
        return
    while start < stop:
        yield start
        start += step


def decompose_b26(index, ascii=False):
    """return a letter (base-26) decomposition of index"""
    if ascii:
        return base_decompose_b26(index)
    return base_decompose_b26(index, "éabcdefghijklmnopqrstuvwxyz")


def get_max_length(eschema, attrname):
    """returns the maximum length allowed for 'attrname'"""
    for cst in eschema.relation_definition(attrname).constraints:
        if isinstance(cst, SizeConstraint) and cst.max:
            return cst.max
    return 300
    # raise AttributeError('No Size constraint on attribute "%s"' % attrname)


_GENERATED_VALUES = {}


class _ValueGenerator:
    """generates integers / dates / strings / etc. to fill a DB table"""

    def __init__(self, eschema, choice_func=None):
        """<choice_func> is a function that returns a list of possible
        choices for a given entity type and an attribute name. It should
        looks like :
            def values_for(etype, attrname):
                # some stuff ...
                return alist_of_acceptable_values # or None
        """
        self.choice_func = choice_func
        self.eschema = eschema

    def generate_attribute_value(self, entity, attrname, index=1, **kwargs):
        if attrname in entity:
            return entity[attrname]
        eschema = self.eschema
        if not eschema.has_unique_values(attrname):
            value = self.__generate_value(entity, attrname, index, **kwargs)
        else:
            value = self.__generate_value(entity, attrname, index, **kwargs)
            while value in _GENERATED_VALUES.get((eschema, attrname), ()):
                index += 1
                value = self.__generate_value(entity, attrname, index, **kwargs)
            _GENERATED_VALUES.setdefault((eschema, attrname), set()).add(value)
        entity[attrname] = value
        return value

    def __generate_value(self, entity, attrname, index, **kwargs):
        """generates a consistent value for 'attrname'"""
        eschema = self.eschema
        attrtype = str(eschema.destination(attrname)).lower()
        # Before calling generate_%s functions, try to find values domain
        if self.choice_func is not None:
            values_domain = self.choice_func(eschema, attrname)
            if values_domain is not None:
                return choice(values_domain)
        gen_func = getattr(
            self,
            f"generate_{eschema}_{attrname}",
            getattr(self, f"generate_Any_{attrname}", None),
        )
        if gen_func is not None:
            return gen_func(entity, index, **kwargs)
        # If no specific values domain, then generate a dummy value
        gen_func = getattr(self, f"generate_{attrtype}")
        return gen_func(entity, attrname, index, **kwargs)

    def generate_string(self, entity, attrname, index, format=None):
        """generates a consistent value for 'attrname' if it's a string"""
        # First try to get choices
        choosed = self.get_choice(entity, attrname)
        if choosed is not None:
            return choosed
        # All other case, generate a default string
        attrlength = get_max_length(self.eschema, attrname)
        num_len = numlen(index)
        if num_len >= attrlength:
            ascii = getattr(self.eschema.rdef(attrname), "internationalizable", False)
            return ("&" + decompose_b26(index, ascii))[:attrlength]
        # always use plain text when no format is specified
        attrprefix = attrname[: max(attrlength - num_len - 1, 0)]
        if format == "text/html":
            value = "<span>é%s<b>%d</b></span>" % (attrprefix, index)
        elif format == "text/rest":
            value = """
title
-----

* %s
* %d
* é&
""" % (
                attrprefix,
                index,
            )
        else:
            value = "é&%s%d" % (attrprefix, index)
        return value[:attrlength]

    def generate_password(self, entity, attrname, index):
        """generates a consistent value for 'attrname' if it's a password"""
        return "toto"

    def generate_integer(self, entity, attrname, index):
        """generates a consistent value for 'attrname' if it's an integer"""
        return self._constrained_generate(entity, attrname, 0, 1, index)

    generate_int = generate_bigint = generate_integer

    def generate_float(self, entity, attrname, index):
        """generates a consistent value for 'attrname' if it's a float"""
        return self._constrained_generate(entity, attrname, 0.0, 1.0, index)

    def generate_decimal(self, entity, attrname, index):
        """generates a consistent value for 'attrname' if it's a float"""
        return Decimal(str(self.generate_float(entity, attrname, index)))

    def generate_datetime(self, entity, attrname, index):
        """generates a random date (format is 'yyyy-mm-dd HH:MM')"""
        base = datetime(
            randint(2000, 2004), randint(1, 12), randint(1, 28), 11, index % 60
        )
        return self._constrained_generate(
            entity, attrname, base, timedelta(hours=1), index
        )

    def generate_tzdatetime(self, entity, attrname, index):
        """generates a random date (format is 'yyyy-mm-dd HH:MM')"""
        base = datetime(
            randint(2000, 2004),
            randint(1, 12),
            randint(1, 28),
            11,
            index % 60,
            tzinfo=timezone.utc,
        )
        return self._constrained_generate(
            entity, attrname, base, timedelta(hours=1), index
        )

    def generate_date(self, entity, attrname, index):
        """generates a random date (format is 'yyyy-mm-dd')"""
        base = date(randint(2000, 2010), 1, 1) + timedelta(randint(1, 365))
        return self._constrained_generate(
            entity, attrname, base, timedelta(days=1), index
        )

    def generate_interval(self, entity, attrname, index):
        """generates a random date (format is 'yyyy-mm-dd')"""
        base = timedelta(randint(1, 365))
        return self._constrained_generate(
            entity, attrname, base, timedelta(days=1), index
        )

    def generate_time(self, entity, attrname, index):
        """generates a random time (format is ' HH:MM')"""
        return time(11, index % 60)  # '11:%02d' % (index % 60)

    generate_tztime = generate_time  # XXX implementation should add a timezone

    def generate_bytes(self, entity, attrname, index, format=None):
        fakefile = Binary(f"{attrname}{index}".encode("ascii"))
        fakefile.filename = f"file_{attrname}"
        return fakefile

    def generate_boolean(self, entity, attrname, index):
        """generates a consistent value for 'attrname' if it's a boolean"""
        return index % 2 == 0

    def _constrained_generate(self, entity, attrname, base, step, index):
        choosed = self.get_choice(entity, attrname)
        if choosed is not None:
            return choosed
        # ensure index > 0
        index += 1
        minvalue, maxvalue = self.get_bounds(entity, attrname)
        if maxvalue is None:
            if minvalue is not None:
                base = max(minvalue, base)
            maxvalue = base + index * step
        if minvalue is None:
            minvalue = maxvalue - (index * step)  # i.e. randint(-index, 0)
        return choice(list(custom_range(minvalue, maxvalue, step)))

    def _actual_boundary(self, entity, attrname, boundary):
        if isinstance(boundary, Attribute):
            # ensure we've a value for this attribute
            entity[attrname] = None  # infinite loop safety belt
            if boundary.attr not in entity:
                self.generate_attribute_value(entity, boundary.attr)
            boundary = actual_value(boundary, entity)
        return boundary

    def get_bounds(self, entity, attrname):
        minvalue = maxvalue = None
        for cst in self.eschema.relation_definition(attrname).constraints:
            if isinstance(cst, IntervalBoundConstraint):
                if cst.minvalue is not None:
                    minvalue = self._actual_boundary(entity, attrname, cst.minvalue)
                if cst.maxvalue is not None:
                    maxvalue = self._actual_boundary(entity, attrname, cst.maxvalue)
            elif isinstance(cst, BoundaryConstraint):
                if cst.operator[0] == "<":
                    maxvalue = self._actual_boundary(entity, attrname, cst.boundary)
                else:
                    minvalue = self._actual_boundary(entity, attrname, cst.boundary)
        return minvalue, maxvalue

    def get_choice(self, entity, attrname):
        """generates a consistent value for 'attrname' if it has some static
        vocabulary set, else return None.
        """
        for cst in self.eschema.relation_definition(attrname).constraints:
            if isinstance(cst, StaticVocabularyConstraint):
                return choice(cst.vocabulary())
        return None

    # XXX nothing to do here
    def generate_Any_data_format(self, entity, index, **kwargs):
        # data_format attribute of File has no vocabulary constraint, we
        # need this method else stupid values will be set which make mtconverter
        # raise exception
        return "application/octet-stream"

    def generate_Any_content_format(self, entity, index, **kwargs):
        # content_format attribute of EmailPart has no vocabulary constraint, we
        # need this method else stupid values will be set which make mtconverter
        # raise exception
        return "text/plain"

    def generate_CWDataImport_log(self, entity, index, **kwargs):
        # content_format attribute of EmailPart has no vocabulary constraint, we
        # need this method else stupid values will be set which make mtconverter
        # raise exception
        logs = [
            "%s\t%s\t%s\t%s<br/>"
            % (
                logging.ERROR,
                "http://url.com?arg1=hop&arg2=hip",
                1,
                xml_escape('hjoio&oio"'),
            )
        ]
        return "<br/>".join(logs)


class autoextend(type):
    def __new__(mcs, name, bases, classdict):
        for attrname, attrvalue in classdict.items():
            if callable(attrvalue):
                if (
                    attrname.startswith("generate_")
                    and len(inspect.getfullargspec(attrvalue).args) < 2
                ):
                    raise TypeError("generate_xxx must accept at least 1 argument")
                setattr(_ValueGenerator, attrname, attrvalue)
        return type.__new__(mcs, name, bases, classdict)


class ValueGenerator(_ValueGenerator, metaclass=autoextend):
    pass


def _default_choice_func(etype, attrname):
    """default choice_func for insert_entity_queries"""
    return None


def insert_entity_queries(
    etype, schema, vreg, entity_num, choice_func=_default_choice_func
):
    """returns a list of 'add entity' queries (couples query, args)
    :type etype: str
    :param etype: the entity's type

    :type schema: cubicweb.schema.Schema
    :param schema: the instance schema

    :type entity_num: int
    :param entity_num: the number of entities to insert

    XXX FIXME: choice_func is here for *historical* reasons, it should
               probably replaced by a nicer way to specify choices
    :type choice_func: function
    :param choice_func: a function that takes an entity type, an attrname and
                        returns acceptable values for this attribute
    """
    queries = []
    for index in range(entity_num):
        restrictions = []
        args = {}
        for attrname, value in make_entity(
            etype, schema, vreg, index, choice_func
        ).items():
            restrictions.append(f"X {attrname} %({attrname})s")
            args[attrname] = value
        if restrictions:
            queries.append((f"INSERT {etype} X: {', '.join(restrictions)}", args))
            assert "eid" not in args, args
        else:
            queries.append((f"INSERT {etype} X", {}))
    return queries


def make_entity(
    etype, schema, vreg, index=0, choice_func=_default_choice_func, form=False
):
    """generates a random entity and returns it as a dict

    by default, generate an entity to be inserted in the repository
    elif form, generate an form dictionary to be given to a web controller
    """
    eschema = schema.entity_schema_for(etype)
    valgen = ValueGenerator(eschema, choice_func)
    entity = attrdict()
    # preprocessing to deal with _format fields
    attributes = []
    relatedfields = {}
    for rschema, attrschema in eschema.attribute_definitions():
        attrname = rschema.type
        if attrname == "eid":
            # don't specify eids !
            continue
        if (
            attrname.endswith("_format")
            and attrname[:-7] in eschema.subject_relations.values()
        ):
            relatedfields[attrname[:-7]] = attrschema
        else:
            attributes.append((attrname, attrschema))
    for attrname, attrschema in attributes:
        if attrname in relatedfields:
            # first generate a format and record it
            format = valgen.generate_attribute_value(
                entity, attrname + "_format", index
            )
            # then a value coherent with this format
            value = valgen.generate_attribute_value(
                entity, attrname, index, format=format
            )
        else:
            value = valgen.generate_attribute_value(entity, attrname, index)
        if form:  # need to encode values
            if attrschema.type == "Bytes":
                # twisted way
                fakefile = value
                filename = value.filename
                value = (filename, "text/plain", fakefile)
            elif attrschema.type == "Date":
                value = value.strftime(vreg.property_value("ui.date-format"))
            elif attrschema.type == "Datetime":
                value = value.strftime(vreg.property_value("ui.datetime-format"))
            elif attrschema.type == "Time":
                value = value.strftime(vreg.property_value("ui.time-format"))
            elif attrschema.type == "Float":
                fmt = vreg.property_value("ui.float-format")
                value = fmt % value
            else:
                value = value
    return entity


def select(constraints, cnx, selectvar="O", objtype=None):
    """returns list of eids matching <constraints>

    <selectvar> should be either 'O' or 'S' to match schema definitions
    """
    try:
        rql = f"Any {selectvar} WHERE {constraints}"
        if objtype:
            rql += f", {selectvar} is {objtype}"
        rset = cnx.execute(rql)
    except Exception:
        print(f"could restrict eid_list with given constraints ({constraints!r})")
        return []
    return set(eid for eid, in rset.rows)


def make_relations_queries(schema, edict, cnx, ignored_relations=(), existingrels=None):
    """returns a list of generated RQL queries for relations
    :param schema: The instance schema

    :param e_dict: mapping between etypes and eids

    :param ignored_relations: list of relations to ignore (i.e. don't try
                              to generate insert queries for these relations)
    """
    gen = RelationsQueriesGenerator(schema, cnx, existingrels)
    return gen.compute_queries(edict, ignored_relations)


def composite_relation(rschema):
    for obj in rschema.objects():
        if (
            obj.relation_definition(rschema, "object", take_first=True).composite
            == "subject"
        ):
            return True
    for obj in rschema.subjects():
        if (
            obj.relation_definition(rschema, "subject", take_first=True).composite
            == "object"
        ):
            return True
    return False


class RelationsQueriesGenerator:
    rql_tmpl = "SET S %s O WHERE S eid %%(subjeid)s, O eid %%(objeid)s"

    def __init__(self, schema, cnx, existing=None):
        self.schema = schema
        self.cnx = cnx
        self.existingrels = existing or {}

    def compute_queries(self, entities_dict, ignored_relations):
        """
        entities_dict: Dict[str(EntityTypeName), Set[Entity]]
        """
        #   1/ skip final relations and explictly ignored relations
        rels = sorted(
            [
                rschema
                for rschema in self.schema.relations()
                if not (rschema.final or rschema.rule or rschema in ignored_relations)
            ],
            key=lambda x: not composite_relation(x),
        )
        # for each relation
        #   2/ take each possible couple (subj, obj)
        #   3/ analyze cardinality of relation
        #      a/ if relation is mandatory, insert one relation
        #      b/ else insert N relations where N is the mininum
        #         of 20 and the number of existing targetable entities
        for rschema in rels:
            sym = set()
            sedict = deepcopy(entities_dict)
            oedict = deepcopy(entities_dict)
            delayed = []
            # for each couple (subjschema, objschema), insert relations
            for subj, obj in rschema.relation_definitions:
                sym.add((subj, obj))
                if rschema.symmetric and (obj, subj) in sym:
                    continue
                subjcard, objcard = rschema.relation_definition(subj, obj).cardinality
                # process mandatory relations first
                if subjcard in "1+" or objcard in "1+" or composite_relation(rschema):
                    for query, args in self.make_relation_queries(
                        sedict, oedict, rschema, subj, obj
                    ):
                        yield query, args
                else:
                    delayed.append((subj, obj))
            for subj, obj in delayed:
                for query, args in self.make_relation_queries(
                    sedict, oedict, rschema, subj, obj
                ):
                    yield query, args

    def qargs(self, subjeids, objeids, subjcard, objcard, subjeid, objeid):
        if subjcard in "?1+":
            subjeids.remove(subjeid)
        if objcard in "?1+":
            objeids.remove(objeid)
        return {"subjeid": subjeid, "objeid": objeid}

    def make_relation_queries(self, sedict, oedict, rschema, subj, obj):
        rdef = rschema.relation_definition(subj, obj)
        subjcard, objcard = rdef.cardinality
        subjeids = sedict.get(subj, frozenset())
        used = self.existingrels[rschema.type]
        preexisting_subjrels = set(subj for subj, obj in used)
        preexisting_objrels = set(obj for subj, obj in used)
        # if there are constraints, only select appropriate objeids
        rql = self.rql_tmpl % rschema.type
        constraints = [c for c in rdef.constraints if isinstance(c, RQLConstraint)]
        if constraints:
            restrictions = ", ".join(c.expression for c in constraints)
            rql += f", {restrictions}"
            # restrict object eids if possible
            # XXX the attempt to restrict below in completely wrong
            # disabling it for now
            objeids = select(restrictions, self.cnx, objtype=obj)
        else:
            objeids = oedict.get(obj, frozenset())
        if subjcard in "?1" or objcard in "?1":
            for subjeid, objeid in used:
                if subjcard in "?1" and subjeid in subjeids:
                    subjeids.remove(subjeid)
                    # XXX why?
                    # if objeid in objeids:
                    #    objeids.remove(objeid)
                if objcard in "?1" and objeid in objeids:
                    objeids.remove(objeid)
                    # XXX why?
                    # if subjeid in subjeids:
                    #    subjeids.remove(subjeid)
        if not subjeids:
            check_card_satisfied(objcard, objeids, subj, rschema, obj)
            return
        if not objeids:
            check_card_satisfied(subjcard, subjeids, subj, rschema, obj)
            return
        if subjcard in "?1+":
            for subjeid in set(subjeids):
                # do not insert relation if this entity already has a relation
                if subjeid in preexisting_subjrels:
                    continue
                objeid = choose_eid(objeids, subjeid)
                if objeid is None or (subjeid, objeid) in used:
                    continue
                yield rql, self.qargs(
                    subjeids, objeids, subjcard, objcard, subjeid, objeid
                )
                used.add((subjeid, objeid))
                if not objeids:
                    check_card_satisfied(subjcard, subjeids, subj, rschema, obj)
                    break
        elif objcard in "?1+":
            for objeid in set(objeids):
                # do not insert relation if this entity already has a relation
                if objeid in preexisting_objrels:
                    continue
                subjeid = choose_eid(subjeids, objeid)
                if subjeid is None or (subjeid, objeid) in used:
                    continue
                yield rql, self.qargs(
                    subjeids, objeids, subjcard, objcard, subjeid, objeid
                )
                used.add((subjeid, objeid))
                if not subjeids:
                    check_card_satisfied(objcard, objeids, subj, rschema, obj)
                    break
        else:
            pairs = list(itertools.product(subjeids, objeids))
            take_only = min(len(pairs), 20)  # FIXME: 20 should be read from config
            for subjeid, objeid in sample(pairs, take_only):
                if subjeid != objeid and (subjeid, objeid) not in used:
                    used.add((subjeid, objeid))
                    yield rql, self.qargs(
                        subjeids, objeids, subjcard, objcard, subjeid, objeid
                    )


def check_card_satisfied(card, remaining, subj, rschema, obj):
    if card in "1+" and remaining:
        raise Exception(
            f"can't satisfy cardinality {card} for relation {subj} {rschema} {obj}"
        )


def choose_eid(values, avoid):
    values = tuple(values)
    if len(values) == 1 and values[0] == avoid:
        return None
    objeid = choice(values)
    while objeid == avoid:  # avoid infinite recursion like in X comment X
        objeid = choice(values)
    return objeid


# UTILITIES FUNCS ##############################################################
def make_tel(num_tel):
    """takes an integer, converts is as a string and inserts
    white spaces each 2 chars (french notation)
    """
    num_list = list(str(num_tel))
    for index in (6, 4, 2):
        num_list.insert(index, " ")

    return "".join(num_list)


def numlen(number):
    """returns the number's length"""
    return len(str(number))



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""the cubicweb-ctl tool, based on logilab.common.clcommands to
provide a pluggable commands system.
"""
# *ctl module should limit the number of import to be imported as quickly as
# possible (for cubicweb-ctl reactivity, necessary for instance for usable bash
# completion). So import locally in command helpers.
import os
import sys
import traceback
from os import listdir
from os.path import exists, join, isdir
from warnings import filterwarnings

from os import kill


from logilab.common.clcommands import CommandLine
from logilab.common.shellutils import ASK
from logilab.common.configuration import merge_options
from logilab.common.decorators import clear_cache

from cubicweb import ConfigurationError, ExecutionError, BadCommandUsage, utils, repoapi
from cubicweb.cwconfig import (
    CubicWebConfiguration as cwcfg,
    instance_configuration,
)
from cubicweb.server import set_debug
from cubicweb.server.repository import Repository, get_cnxset
from cubicweb.toolsutils import Command, rm, create_dir, underline_title
from cubicweb.__pkginfo__ import version as cw_version

LOG_LEVELS = ("debug", "info", "warning", "error")
DBG_FLAGS = ("RQL", "SQL", "REPO", "HOOKS", "OPS", "SEC", "MORE", "ALL")

# don't check duplicated commands, it occurs when reloading site_cubicweb
CWCTL = CommandLine(
    "cubicweb-ctl",
    "The CubicWeb swiss-knife.",
    version=cw_version,
    check_duplicated_command=False,
)


def wait_process_end(pid, maxtry=10, waittime=1):
    """wait for a process to actually die"""
    import signal
    from time import sleep

    nbtry = 0
    while nbtry < maxtry:
        kill(pid, signal.SIGUSR1)
        nbtry += 1
        sleep(waittime)
    else:
        raise ExecutionError(f"can't kill process {pid}")


def list_instances(regdir):
    if isdir(regdir):
        return sorted(idir for idir in listdir(regdir) if isdir(join(regdir, idir)))
    else:
        return []


def detect_available_modes(templdir):
    modes = []
    for fname in ("schema", "schema.py"):
        if exists(join(templdir, fname)):
            modes.append("repository")
            break
    for fname in ("data", "views", "views.py"):
        if exists(join(templdir, fname)):
            modes.append("web ui")
            break
    return modes


def available_cube_names(cwcfg):
    """Return a list of available cube names, with 'cubicweb_' prefix dropped."""

    def drop_prefix(cube):
        prefix = "cubicweb_"
        if cube.startswith(prefix):
            cube = cube[len(prefix) :]
        return cube

    return [drop_prefix(cube) for cube in cwcfg.available_cubes()]


class InstanceCommand(Command):
    """base class for command taking one instance id as arguments"""

    arguments = "<instance>"

    # enforce having one instance
    min_args = 0
    max_args = 1

    options = (
        (
            "force",
            {
                "short": "f",
                "action": "store_true",
                "default": False,
                "help": "force command without asking confirmation",
            },
        ),
        (
            "pdb",
            {
                "action": "store_true",
                "default": False,
                "help": "launch pdb on exception",
            },
        ),
        (
            "loglevel",
            {
                "type": "choice",
                "default": None,
                "metavar": "<log level>",
                "choices": LOG_LEVELS,
                "short": "l",
                "help": "allow to specify log level for debugging (choices: %s)"
                % (", ".join(LOG_LEVELS)),
            },
        ),
        (
            "dbglevel",
            {
                "type": "multiple_choice",
                "metavar": "<debug level>",
                "default": None,
                "choices": DBG_FLAGS,
                "help": (
                    "Set the server debugging flags; you may choose several "
                    'values in %s; imply "debug" loglevel if loglevel is not set'
                    % (DBG_FLAGS,)
                ),
            },
        ),
    )
    actionverb = None

    @property
    def _cnx(self):
        return repoapi.get_repository(config=self.cwconfig).internal_cnx()

    def run(self, args):
        """run the <command>_method on each argument (a list of instance
        identifiers)
        """
        if not args:
            if "CW_INSTANCE" in os.environ:
                appid = os.environ["CW_INSTANCE"]
            else:
                raise BadCommandUsage("Error: instance id is missing")
        else:
            appid = args[0]

        cmdmeth = getattr(self, f"{self.name}_instance")

        traceback_ = None

        # debugmode=True is to force to have a StreamHandler used instead of
        # writting the logs into a file in /tmp
        self.cwconfig = cwcfg.config_for(appid, debugmode=True)

        # by default loglevel is 'error' but we keep the default value to None
        # because some subcommands (e.g: pyramid) can override the loglevel in
        # certain situations if it's not explicitly set by the user and we want
        # to detect that (the "None" case)
        if self["loglevel"] is None:
            # if no loglevel is set but dbglevel is here we want to set level to debug
            if self["dbglevel"]:
                init_cmdline_log_threshold(self.cwconfig, "debug")
            else:
                init_cmdline_log_threshold(self.cwconfig, "error")
        else:
            init_cmdline_log_threshold(self.cwconfig, self["loglevel"])

        if self["dbglevel"]:
            set_debug("|".join("DBG_" + x.upper() for x in self["dbglevel"]))

        try:
            status = cmdmeth(appid) or 0
        except (ExecutionError, ConfigurationError) as ex:
            # we need to do extract this information here for pdb since it is
            # now lost in python 3 once we exit the try/catch statement
            exception_type, exception, traceback_ = sys.exc_info()

            sys.stderr.write(f"instance {appid} not {self.actionverb}: {ex}\n")
            status = 4
        except Exception as ex:
            # idem
            exception_type, exception, traceback_ = sys.exc_info()

            traceback.print_exc()

            sys.stderr.write(f"instance {appid} not {self.actionverb}: {ex}\n")
            status = 8

        except (KeyboardInterrupt, SystemExit) as ex:
            # idem
            exception_type, exception, traceback_ = sys.exc_info()

            sys.stderr.write(f"{self.name} aborted\n")
            if isinstance(ex, KeyboardInterrupt):
                status = 2  # specific error code
            else:
                status = ex.code

        if status != 0 and self.config.pdb:
            pdb = utils.get_pdb()

            if traceback_ is not None:
                pdb.post_mortem(traceback_)
            else:
                print(
                    "WARNING: Could not access to the traceback because the command return "
                    "code is different than 0 but the command didn't raised an exception."
                )
                # we can't use "header=" of set_trace because ipdb doesn't supports it
                pdb.set_trace()

        sys.exit(status)


# base commands ###############################################################


class ListCommand(Command):
    """List configurations, cubes and instances.

    List available installed cubes and registered instances.

    If given, the optional argument allows to restrict listing only a category of items.
    """

    name = "list"
    arguments = "[all|cubes|instances]"
    options = (
        (
            "verbose",
            {"short": "v", "action": "store_true", "help": "display more information."},
        ),
    )

    def run(self, args):
        """run the command with its specific arguments"""
        if not args:
            mode = "all"
        elif len(args) == 1:
            mode = args[0]
        else:
            raise BadCommandUsage("Too many arguments")

        from cubicweb.migration import ConfigurationProblem

        if mode == "all":
            print(f"CubicWeb {cwcfg.cubicweb_version()} ({cwcfg.mode} mode)")
            print()

        if mode in ("all", "cubes"):
            cfgpb = ConfigurationProblem(cwcfg)
            try:
                cube_names = available_cube_names(cwcfg)
                namesize = max(len(x) for x in cube_names)
            except ConfigurationError as ex:
                print("No cubes available:", ex)
            except ValueError:
                print("No cubes available")
            else:
                print("Available cubes:")
                for cube in cube_names:
                    try:
                        tinfo = cwcfg.cube_pkginfo(cube)
                        tversion = tinfo.version
                        cfgpb.add_cube(cube, tversion)
                    except (ConfigurationError, AttributeError) as ex:
                        tinfo = None
                        tversion = f"[missing cube information: {ex}]"
                    print(f"* {cube.ljust(namesize)} {tversion}")
                    if self.config.verbose:
                        if tinfo:
                            descr = getattr(tinfo, "description", "")
                            if not descr:
                                descr = tinfo.__doc__
                            if descr:
                                print("    " + "    \n".join(descr.splitlines()))
                        modes = detect_available_modes(cwcfg.cube_dir(cube))
                        print(f"    available modes: {', '.join(modes)}")
            print()

        if mode in ("all", "instances"):
            try:
                regdir = cwcfg.instances_dir()
            except ConfigurationError as ex:
                print("No instance available:", ex)
                print()
                return
            instances = list_instances(regdir)
            if instances:
                print(f"Available instances ({regdir}):")
                for appid in instances:
                    print(f"* {appid}")
            else:
                print(f"No instance available in {regdir}")
            print()

        if mode == "all":
            # configuration management problem solving
            cfgpb.solve()
            if cfgpb.warnings:
                print("Warnings:\n", "\n".join("* " + txt for txt in cfgpb.warnings))
            if cfgpb.errors:
                print("Errors:")
                for op, cube, version, src in cfgpb.errors:
                    if op == "add":
                        print("* cube", cube, end=" ")
                        if version:
                            print(" version", version, end=" ")
                        print(f"is not installed, but required by {src}")
                    else:
                        print(
                            "* cube %s version %s is installed, but version %s is required by %s"
                            % (cube, cfgpb.cubes[cube], version, src)
                        )


def check_options_consistency(config):
    if config.automatic and config.config_level > 0:
        raise BadCommandUsage(
            "--automatic and --config-level should not be " "used together"
        )


class CreateInstanceCommand(Command):
    """Create an instance from a cube. This is a unified
    command which can handle web / server / all-in-one installation
    according to available parts of the software library and of the
    desired cube.

    <cube>
      the name of cube to use (list available cube names using
      the "list" command). You can use several cubes by separating
      them using comma (e.g. 'jpl,email')
    <instance>
      an identifier for the instance to create
    """

    name = "create"
    arguments = "<cube> <instance>"
    min_args = max_args = 2
    options = (
        (
            "automatic",
            {
                "short": "a",
                "action": "store_true",
                "default": False,
                "help": (
                    "automatic mode: never ask and use default answer to every "
                    "question. this may require that your login match a database super "
                    "user (allowed to create database & all)."
                ),
            },
        ),
        (
            "config-level",
            {
                "short": "l",
                "type": "int",
                "metavar": "<level>",
                "default": 0,
                "help": (
                    "configuration level (0..2): 0 will ask for essential "
                    "configuration parameters only while 2 will ask for all parameters"
                ),
            },
        ),
        (
            "no-db-create",
            {
                "short": "S",
                "action": "store_true",
                "default": False,
                "help": "stop after creation and do not continue with db-create",
            },
        ),
    )

    def run(self, args):
        """run the command with its specific arguments"""
        from logilab.common.textutils import splitstrip

        check_options_consistency(self.config)
        configname = "all-in-one"
        cubes, appid = args
        cubes = splitstrip(cubes)
        # get the configuration and helper
        config = cwcfg.config_for(appid, configname, creating=True)
        cubes = config.expand_cubes(cubes)
        config.init_cubes(cubes)
        helper = self.config_helper(config)
        # check the cube exists
        try:
            templdirs = [cwcfg.cube_dir(cube) for cube in cubes]
        except ConfigurationError as ex:
            print(ex)
            print("\navailable cubes:", end=" ")
            print(", ".join(available_cube_names(cwcfg)))
            return
        # create the registry directory for this instance
        print("\n" + underline_title(f"Creating the instance {appid}"))
        create_dir(config.apphome)
        # cubicweb-ctl configuration
        if not self.config.automatic:
            print(
                "\n" + underline_title(f"Configuring the instance ({configname}.conf)")
            )
            config.input_config("main", self.config.config_level)
        # configuration'specific stuff
        print()
        helper.bootstrap(cubes, self.config.automatic, self.config.config_level)
        # input for cubes specific options
        if not self.config.automatic:
            sections = set(
                sect.lower()
                for sect, opt, odict in config.all_options()
                if "type" in odict and odict.get("level", 0) <= self.config.config_level
            )
            for section in sections:
                if section not in ("main", "email", "web"):
                    print("\n" + underline_title(f"{section} options"))
                    config.input_config(section, self.config.config_level)
        # write down configuration
        config.save()
        print(f"-> generated config {config.main_config_file()}")
        # handle i18n files structure
        # in the first cube given
        from cubicweb import i18n

        langs = [
            lang for lang, _ in i18n.available_catalogs(join(templdirs[0], "i18n"))
        ]
        errors = config.i18ncompile(langs)
        if errors:
            print("\n".join(errors))
            if self.config.automatic or not ASK.confirm(
                "error while compiling message catalogs, " "continue anyway ?"
            ):
                print("creation not completed")
                return
        # create the additional data directory for this instance
        if config.appdatahome != config.apphome:  # true in dev mode
            create_dir(config.appdatahome)
        create_dir(join(config.appdatahome, "backup"))
        if config["uid"]:
            from logilab.common.shellutils import chown

            # this directory should be owned by the uid of the server process
            print(f"set {config['uid']} as owner of the data directory")
            chown(config.appdatahome, config["uid"])
        print(f"\n-> creation done for {repr(config.apphome)[1:-1]}\n")
        if not self.config.no_db_create:
            helper.postcreate(self.config.automatic, self.config.config_level)


class DeleteInstanceCommand(Command):
    """Delete an instance. Will remove instance's files and
    unregister it.
    """

    name = "delete"
    arguments = "<instance>"
    min_args = max_args = 1

    def run(self, args):
        """run the command with its specific arguments"""
        appid = args[0]
        config = cwcfg.config_for(appid)
        helper = self.config_helper(config, required=False)
        if helper:
            helper.cleanup()
        # remove home
        rm(config.apphome)
        # remove instance data directory
        try:
            rm(config.appdatahome)
        except OSError as ex:
            import errno

            if ex.errno != errno.ENOENT:
                raise
        print(f"-> instance {appid} deleted.")


# instance commands ########################################################


def init_cmdline_log_threshold(config, loglevel):
    if loglevel is not None:
        config.global_set_option("log-threshold", loglevel.upper())
        config.init_log(config["log-threshold"], force=True)


class UpgradeInstanceCommand(InstanceCommand):
    """Upgrade an instance after cubicweb and/or component(s) upgrade.

    For repository update, you will be prompted for a login / password to use
    to connect to the system database.  For some upgrades, the given user
    should have create or alter table permissions.

    <instance>
      identifier of the instance to upgrade.
    """

    name = "upgrade"
    actionverb = "upgraded"
    options = InstanceCommand.options + (
        (
            "force-cube-version",
            {
                "short": "t",
                "type": "named",
                "metavar": "cube1:X.Y.Z,cube2:X.Y.Z",
                "default": None,
                "help": "force migration from the indicated version for the specified cube(s).",
            },
        ),
        (
            "force-cubicweb-version",
            {
                "short": "e",
                "type": "string",
                "metavar": "X.Y.Z",
                "default": None,
                "help": "force migration from the indicated cubicweb version.",
            },
        ),
        (
            "fs-only",
            {
                "short": "s",
                "action": "store_true",
                "default": False,
                "help": "only upgrade files on the file system, not the database.",
            },
        ),
        (
            "no-config-update",
            {
                "short": "C",
                "action": "store_true",
                "default": False,
                "help": "do NOT update config file if set.",
            },
        ),
        (
            "verbosity",
            {
                "short": "v",
                "type": "int",
                "metavar": "<0..2>",
                "default": 1,
                "help": "0: no confirmation, 1: only main commands confirmed, 2 ask \
for everything.",
            },
        ),
        (
            "backup-db",
            {
                "short": "b",
                "type": "yn",
                "metavar": "<y or n>",
                "default": None,
                "help": (
                    "Backup the instance database before upgrade.\n"
                    "If the option is ommitted, confirmation will be ask."
                ),
            },
        ),
        (
            "ext-sources",
            {
                "short": "E",
                "type": "csv",
                "metavar": "<sources>",
                "default": None,
                "help": "For multisources instances, specify to which sources the \
repository should connect to for upgrading. When unspecified or 'migration' is \
given, appropriate sources for migration will be automatically selected \
(recommended). If 'all' is given, will connect to all defined sources.",
            },
        ),
    )

    def upgrade_instance(self, appid):
        print("\n" + underline_title(f"Upgrading the instance {appid}"))
        from logilab.common.changelog import Version

        config = cwcfg.config_for(appid)
        config.repairing = True  # notice we're not starting the server
        config.verbosity = self.config.verbosity
        set_sources_mode = getattr(config, "set_sources_mode", None)
        if set_sources_mode is not None:
            set_sources_mode(self.config.ext_sources or ("migration",))
        # get instance and installed versions for the server and the componants
        mih = config.migration_handler()
        repo = mih.repo
        vcconf = repo.get_versions()
        helper = self.config_helper(config, required=False)
        if self.config.force_cube_version:
            for cube, version in self.config.force_cube_version.items():
                vcconf[cube] = Version(version)
        toupgrade = []
        for cube in config.cubes():
            installedversion = config.cube_version(cube)
            try:
                applversion = vcconf[cube]
            except KeyError:
                config.error(f"no version information for {cube}")
                continue
            if installedversion > applversion:
                toupgrade.append((cube, applversion, installedversion))
        cubicwebversion = config.cubicweb_version()
        if self.config.force_cubicweb_version:
            applcubicwebversion = Version(self.config.force_cubicweb_version)
            vcconf["cubicweb"] = applcubicwebversion
        else:
            applcubicwebversion = vcconf.get("cubicweb")
        if cubicwebversion > applcubicwebversion:
            toupgrade.append(("cubicweb", applcubicwebversion, cubicwebversion))
        # run cubicweb/componants migration scripts
        if self.config.fs_only or toupgrade:
            for cube, fromversion, toversion in toupgrade:
                print(
                    f"-> migration needed from {fromversion} to {toversion} for {cube}"
                )
            with mih.cnx:
                with mih.cnx.security_enabled(False, False):
                    mih.migrate(vcconf, reversed(toupgrade), self.config)
            clear_cache(config, "instance_md5_version")
        else:
            print(f"-> no data migration needed for instance {appid}.")
        # rewrite main configuration file
        if not self.config.no_config_update:
            mih.rewrite_configuration()
        mih.shutdown()
        # handle i18n upgrade
        if not self.i18nupgrade(config):
            return
        print()
        if helper:
            helper.postupgrade(repo)
        print("-> instance migrated.")
        print()

    def i18nupgrade(self, config):
        # handle i18n upgrade:
        # * install new languages
        # * recompile catalogs
        # XXX search available language in the first cube given
        from cubicweb import i18n

        templdir = cwcfg.cube_dir(config.cubes()[0])
        langs = [lang for lang, _ in i18n.available_catalogs(join(templdir, "i18n"))]
        errors = config.i18ncompile(langs)
        if errors:
            print("\n".join(errors))
            if not ASK.confirm(
                "Error while compiling message catalogs, " "continue anyway?"
            ):
                print("-> migration not completed.")
                return False
        return True


class ListVersionsInstanceCommand(InstanceCommand):
    """List versions used by an instance.

    <instance>...
      identifiers of the instances to list versions for.
    """

    name = "versions"

    def versions_instance(self, appid):
        config = self.cwconfig
        # should not raise error if db versions don't match fs versions
        config.repairing = True
        # no need to load all appobjects and schema
        config.quick_start = True
        if hasattr(config, "set_sources_mode"):
            config.set_sources_mode(("migration",))
        config = instance_configuration(appid)
        # We need to initialize only a few things to have all cube versions.
        # So we can't use regular methods, since we only need a Repository
        # with a configuration, a connection and a schema.
        # Thus, we'll be able to list all cube versions of the instance,
        # even if some cubes aren't installed, which is impossible
        # if we simply use config.repository().get_versions().
        repo = Repository(config)
        repo.cnxsets = get_cnxset(repo.config, repo.system_source, bootstrap=True)
        repo.set_schema(repo.config.load_bootstrap_schema(), resetvreg=False)
        vcconf = repo.get_versions(checkversions=False)
        for key in sorted(vcconf):
            print(key + ": %s.%s.%s" % vcconf[key])


class ShellCommand(Command):
    """Run an interactive migration shell on an instance. This is a python shell
    with enhanced migration commands predefined in the namespace. An additional
    argument may be given corresponding to a file containing commands to execute
    in batch mode.

    By default it will connect to a local instance using an in memory
    connection, unless a URL to a running instance is specified.

    Arguments after bare "--" string will not be processed by the shell command
    You can use it to pass extra arguments to your script and expect for
    them in '__args__' afterwards.

    <instance>
      the identifier of the instance to connect.
    """

    name = "shell"
    arguments = "<instance> [batch command file(s)] [-- <script arguments>]"
    min_args = 1
    max_args = None
    options = merge_options(
        (
            (
                "system-only",
                {
                    "short": "S",
                    "action": "store_true",
                    "help": (
                        "only connect to the system source when the instance is "
                        "using multiple sources. You can't use this option and the "
                        "--ext-sources option at the same time."
                    ),
                    "group": "local",
                },
            ),
            (
                "ext-sources",
                {
                    "short": "E",
                    "type": "csv",
                    "metavar": "<sources>",
                    "help": "For multisources instances, specify to which sources the \
repository should connect to for upgrading. When unspecified or 'all' given, \
will connect to all defined sources. If 'migration' is given, appropriate \
sources for migration will be automatically selected.",
                    "group": "local",
                },
            ),
            (
                "force",
                {
                    "short": "f",
                    "action": "store_true",
                    "help": "don't check instance is up to date.",
                    "group": "local",
                },
            ),
        )
        + InstanceCommand.options
    )

    def _get_mih(self, appid):
        """returns migration context handler & shutdown function"""
        config = cwcfg.config_for(appid)
        if self.config.ext_sources:
            assert not self.config.system_only
            sources = self.config.ext_sources
        elif self.config.system_only:
            sources = ("system",)
        else:
            sources = ("all",)
        config.set_sources_mode(sources)
        config.repairing = self.config.force
        mih = config.migration_handler()
        return mih, lambda: mih.shutdown()

    def run(self, args):
        appuri = args.pop(0)
        mih, shutdown_callback = self._get_mih(appuri)
        try:
            with mih.cnx:
                with mih.cnx.security_enabled(False, False):
                    if args:
                        # use cmdline parser to access left/right attributes only
                        # remember that usage requires instance appid as first argument
                        scripts, args = (
                            self.cmdline_parser.largs[1:],
                            self.cmdline_parser.rargs,
                        )
                        for script in scripts:
                            mih.cmd_process_script(script, scriptargs=args)
                            mih.commit()
                    else:
                        mih.interactive_shell()
        finally:
            shutdown_callback()


class RecompileInstanceCatalogsCommand(InstanceCommand):
    """Recompile i18n catalogs for instances.

    <instance>
      identifier of the instance to consider.
    """

    name = "i18ninstance"

    def i18ninstance_instance(self, appid):
        """recompile instance's messages catalogs"""
        config = self.cwconfig
        config.quick_start = True  # notify this is not a regular start
        repo = config.repository()
        if config._cubes is None:
            # web only config
            config.init_cubes(repo.get_cubes())
        errors = config.i18ncompile()
        if errors:
            print("\n".join(errors))


class ListInstancesCommand(Command):
    """list available instances, useful for bash completion."""

    name = "listinstances"
    hidden = True

    def run(self, args):
        """run the command with its specific arguments"""
        regdir = cwcfg.instances_dir()
        for appid in sorted(listdir(regdir)):
            print(appid)


class ListCubesCommand(Command):
    """list available componants, useful for bash completion."""

    name = "listcubes"
    hidden = True

    def run(self, args):
        """run the command with its specific arguments"""
        for cube in cwcfg.available_cubes():
            print(cube)


class ConfigureInstanceCommand(InstanceCommand):
    """Configure instance.

    <instance>
      identifier of the instance to configure.
    """

    name = "configure"
    actionverb = "configured"

    options = merge_options(
        InstanceCommand.options
        + (
            (
                "param",
                {
                    "short": "p",
                    "type": "named",
                    "metavar": "key1:value1,key2:value2",
                    "default": None,
                    "help": "set <key> to <value> in configuration file.",
                },
            ),
        ),
    )

    def configure_instance(self, appid):
        if self.config.param is not None:
            appcfg = self.cwconfig
            for key, value in self.config.param.items():
                try:
                    appcfg.global_set_option(key, value)
                except KeyError:
                    raise ConfigurationError(
                        f'unknown configuration key "{key}" for mode {appcfg.name}'
                    )
            appcfg.save()


for cmdcls in (
    ListCommand,
    CreateInstanceCommand,
    DeleteInstanceCommand,
    UpgradeInstanceCommand,
    ListVersionsInstanceCommand,
    ShellCommand,
    RecompileInstanceCatalogsCommand,
    ListInstancesCommand,
    ListCubesCommand,
    ConfigureInstanceCommand,
):
    CWCTL.register(cmdcls)


def run(args=sys.argv[1:]):
    """command line tool"""
    filterwarnings("default", category=DeprecationWarning)
    cwcfg.load_cwctl_plugins()
    try:
        CWCTL.run(args)
    except ConfigurationError as err:
        print("ERROR: ", err)
        sys.exit(1)
    except ExecutionError as err:
        print(err)
        sys.exit(2)


if __name__ == "__main__":
    run()



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""schema definition related entities"""


from logilab.common.decorators import cached
from yams.schema import role_name

from yams import ValidationError
from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.schema import ERQLExpression, RRQLExpression


class CWEType(AnyEntity):
    __regid__ = "CWEType"
    fetch_attrs, cw_fetch_order = fetch_config(["name"])

    def dc_title(self):
        return f"{self.name} ({self._cw._(self.name)})"

    def dc_long_title(self):
        stereotypes = []
        _ = self._cw._
        if self.final:
            stereotypes.append(_("final"))
        if stereotypes:
            return f"{self.dc_title()} <<{', '.join(stereotypes)}>>"
        return self.dc_title()


class CWRType(AnyEntity):
    __regid__ = "CWRType"
    fetch_attrs, cw_fetch_order = fetch_config(["name"])

    def dc_title(self):
        return f"{self.name} ({self._cw._(self.name)})"

    def dc_long_title(self):
        stereotypes = []
        _ = self._cw._
        if self.symmetric:
            stereotypes.append(_("symmetric"))
        if self.inlined:
            stereotypes.append(_("inlined"))
        if self.final:
            stereotypes.append(_("final"))
        if stereotypes:
            return f"{self.dc_title()} <<{', '.join(stereotypes)}>>"
        return self.dc_title()

    def check_inlined_allowed(self):
        """check inlining is possible, raise ValidationError if not possible"""
        # don't use the persistent schema, we may miss cardinality changes
        # in the same transaction
        for rdef in self.reverse_relation_type:
            card = rdef.cardinality[0]
            if card not in "?1":
                qname = role_name("inlined", "subject")
                rtype = self.name
                stype = rdef.stype
                otype = rdef.otype
                msg = self._cw._(
                    "can't set inlined=True, "
                    "%(stype)s %(rtype)s %(otype)s "
                    "has cardinality=%(card)s"
                )
                raise ValidationError(self.eid, {qname: msg % locals()})


class CWRelation(AnyEntity):
    __regid__ = "CWRelation"
    fetch_attrs = fetch_config(["cardinality"])[0]

    def dc_title(self):
        return "{} {} {}".format(
            self.from_entity[0].name,
            self.relation_type[0].name,
            self.to_entity[0].name,
        )

    def dc_long_title(self):
        card = self.cardinality
        scard, ocard = "", ""
        if card[0] != "1":
            scard = f"[{card[0]}]"
        if card[1] != "1":
            ocard = f"[{card[1]}]"
        return "{} {}{}{} {}".format(
            self.from_entity[0].name,
            scard,
            self.relation_type[0].name,
            ocard,
            self.to_entity[0].name,
        )

    @property
    def rtype(self):
        return self.relation_type[0]

    @property
    def stype(self):
        return self.from_entity[0]

    @property
    def otype(self):
        return self.to_entity[0]

    def yams_schema(self):
        rschema = self._cw.vreg.schema.relation_schema_for(self.rtype.name)
        return rschema.relation_definitions[(self.stype.name, self.otype.name)]


class CWAttribute(CWRelation):
    __regid__ = "CWAttribute"

    def dc_long_title(self):
        card = self.cardinality
        scard = ""
        if card[0] == "1":
            scard = "+"
        return "{} {}{} {}".format(
            self.from_entity[0].name,
            scard,
            self.relation_type[0].name,
            self.to_entity[0].name,
        )


class CWConstraint(AnyEntity):
    __regid__ = "CWConstraint"
    fetch_attrs, cw_fetch_order = fetch_config(["value"])

    def __repr__(self):
        return "<Entity {} {} of type {} {} at {}>".format(
            self.e_schema,
            self.eid,
            self.type,
            list(self.cw_attr_cache),
            id(self),
        )

    def dc_title(self):
        return f"{self.cstrtype[0].name}({self.value or ''})"

    @property
    def type(self):
        return self.cstrtype[0].name


class RQLExpression(AnyEntity):
    __regid__ = "RQLExpression"
    fetch_attrs, cw_fetch_order = fetch_config(["exprtype", "mainvars", "expression"])

    def dc_title(self):
        return self.expression or ""

    def dc_long_title(self):
        return f"{self.exprtype}({self.expression or ''})"

    @property
    def expression_of(self):
        for rel in (
            "read_permission",
            "add_permission",
            "delete_permission",
            "update_permission",
            "condition",
        ):
            values = getattr(self, f"reverse_{rel}")
            if values:
                return values[0]

    @cached
    def _rqlexpr(self):
        if self.exprtype == "ERQLExpression":
            return ERQLExpression(self.expression, self.mainvars, self.eid)
        # if self.exprtype == 'RRQLExpression':
        return RRQLExpression(self.expression, self.mainvars, self.eid)

    def check_expression(self, *args, **kwargs):
        return self._rqlexpr().check(*args, **kwargs)



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for computed attributes/relations hooks"""

from unittest import TestCase

from yams.buildobjs import EntityType, String, Int, SubjectRelation

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.schema import build_schema_from_namespace


class FormulaDependenciesMatrixTC(TestCase):
    def simple_schema(self):
        THISYEAR = 2014

        class Person(EntityType):
            name = String()
            salary = Int()
            birth_year = Int(required=True)
            works_for = SubjectRelation("Company")
            age = Int(formula="Any %d - D WHERE X birth_year D" % THISYEAR)

        class Company(EntityType):
            name = String()
            total_salary = Int(
                formula="Any SUM(SA) GROUPBY X WHERE P works_for X, P salary SA"
            )

        schema = build_schema_from_namespace(vars().items())
        return schema

    def setUp(self):
        from cubicweb.hooks.synccomputed import _FormulaDependenciesMatrix

        self.schema = self.simple_schema()
        self.dependencies = _FormulaDependenciesMatrix(self.schema)

    def test_computed_attributes_by_etype(self):
        comp_by_etype = self.dependencies.computed_attribute_by_etype
        self.assertEqual(len(comp_by_etype), 2)
        values = comp_by_etype["Person"]
        self.assertEqual(len(values), 1)
        self.assertEqual(values[0].relation_type, "age")
        values = comp_by_etype["Company"]
        self.assertEqual(len(values), 1)
        self.assertEqual(values[0].relation_type, "total_salary")

    def test_computed_attribute_by_relation(self):
        comp_by_rdef = self.dependencies.computed_attribute_by_relation
        self.assertEqual(len(comp_by_rdef), 1)
        key, values = next(iter(comp_by_rdef.items()))
        self.assertEqual(key.relation_type, "works_for")
        self.assertEqual(len(values), 1)
        self.assertEqual(values[0].relation_type, "total_salary")

    def test_computed_attribute_by_etype_attrs(self):
        comp_by_attr = self.dependencies.computed_attribute_by_etype_attrs
        self.assertEqual(len(comp_by_attr), 1)
        values = comp_by_attr["Person"]
        self.assertEqual(len(values), 2)
        values = {(rdef.formula, tuple(v)) for rdef, v in values.items()}
        self.assertEqual(
            values,
            {
                ("Any 2014 - D WHERE X birth_year D", tuple(("birth_year",))),
                (
                    "Any SUM(SA) GROUPBY X WHERE P works_for X, P salary SA",
                    tuple(("salary",)),
                ),
            },
        )


class ComputedAttributeTC(CubicWebTC):
    appid = "data-computed"

    def setup_entities(self, cnx):
        self.societe = cnx.create_entity("Societe", nom="Foo")
        cnx.create_entity(
            "Person", name="Titi", salaire=1000, travaille=self.societe, birth_year=2001
        )
        self.tata = cnx.create_entity(
            "Person", name="Tata", salaire=2000, travaille=self.societe, birth_year=1990
        )

    def test_update_on_add_remove_relation(self):
        """check the rewriting of a computed attribute"""
        with self.admin_access.cnx() as cnx:
            self.setup_entities(cnx)
            cnx.commit()
            rset = cnx.execute('Any S WHERE X salaire_total S, X nom "Foo"')
            self.assertEqual(rset[0][0], 3000)
            # Add relation.
            toto = cnx.create_entity(
                "Person",
                name="Toto",
                salaire=1500,
                travaille=self.societe,
                birth_year=1988,
            )
            cnx.commit()
            rset = cnx.execute('Any S WHERE X salaire_total S, X nom "Foo"')
            self.assertEqual(rset[0][0], 4500)
            # Delete relation.
            toto.cw_set(travaille=None)
            cnx.commit()
            rset = cnx.execute('Any S WHERE X salaire_total S, X nom "Foo"')
            self.assertEqual(rset[0][0], 3000)

    def test_recompute_on_attribute_update(self):
        """check the modification of an attribute triggers the update of the
        computed attributes that depend on it"""
        with self.admin_access.cnx() as cnx:
            self.setup_entities(cnx)
            cnx.commit()
            rset = cnx.execute('Any S WHERE X salaire_total S, X nom "Foo"')
            self.assertEqual(rset[0][0], 3000)
            # Update attribute.
            self.tata.cw_set(salaire=1000)
            cnx.commit()
            rset = cnx.execute('Any S WHERE X salaire_total S, X nom "Foo"')
            self.assertEqual(rset[0][0], 2000)

    def test_init_on_entity_creation(self):
        """check the computed attribute is initialized on entity creation"""
        with self.admin_access.cnx() as cnx:
            cnx.create_entity("Person", name="Tata", salaire=2000, birth_year=1990)
            cnx.commit()
            rset = cnx.execute('Any A, X WHERE X age A, X name "Tata"')
            self.assertEqual(rset[0][0], 2014 - 1990)

    def test_recompute_on_ambiguous_relation(self):
        # check we don't end up with TypeResolverException as in #4901163
        with self.admin_access.client_cnx() as cnx:
            societe = cnx.create_entity("Societe", nom="Foo")
            cnx.create_entity("MirrorEntity", mirror_of=societe, extid="1")
            cnx.commit()


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
import logging
from os.path import abspath, join, dirname

from logilab.common.umessage import message_from_string

from cubicweb import ExecutionError
from cubicweb.devtools import (
    DEFAULT_PSQL_SOURCES,
    DEFAULT_SOURCES,
    BASE_URL,
)
from cubicweb.pyramid.config import AllInOneConfiguration


# email handling, to test emails sent by an application ########################

MAILBOX = []


class Email:
    """you'll get instances of Email into MAILBOX during tests that trigger
    some notification.

    * `msg` is the original message object

    * `recipients` is a list of email address which are the recipients of this
      message
    """

    def __init__(self, fromaddr, recipients, msg):
        self.fromaddr = fromaddr
        self.recipients = recipients
        self.msg = msg

    @property
    def message(self):
        return message_from_string(self.msg)

    @property
    def subject(self):
        return self.message.get("Subject")

    @property
    def content(self):
        return self.message.get_payload(decode=True)

    def __repr__(self):
        return "<Email to %s with subject %s>" % (
            ",".join(self.recipients),
            self.message.get("Subject"),
        )


# the trick to get email into MAILBOX instead of actually sent: monkey patch
# cwconfig.SMTP object
class MockSMTP:
    def __init__(self, server, port):
        pass

    def close(self):
        pass

    def sendmail(self, fromaddr, recipients, msg):
        MAILBOX.append(Email(fromaddr, recipients, msg.decode("utf-8")))


class ApptestConfiguration(AllInOneConfiguration):
    name = "all-in-one"  # so it search for all-in-one.conf, not repository.conf
    mode = "test"
    read_instance_schema = False
    init_repository = True
    skip_db_create_and_restore = False
    default_sources = DEFAULT_SOURCES
    smtp_cls = MockSMTP

    def available_languages(self, *args):
        return self.cw_languages()

    def __init__(self, appid, test_module_file, log_threshold=logging.CRITICAL + 10):
        # must be set before calling parent __init__
        apphome = abspath(join(dirname(test_module_file), appid))
        self._apphome = apphome
        super().__init__(appid)
        self.init_log(log_threshold, force=True)
        # need this, usually triggered by cubicweb-ctl
        self.load_cwctl_plugins()
        self.test_module_file = test_module_file

    # By default anonymous login are allow but some test need to deny of to
    # change the default user. Set it to None to prevent anonymous login.
    anonymous_credential = ("anon", "anon")

    def anonymous_user(self):
        if not self.anonymous_credential:
            return None, None
        return self.anonymous_credential

    def set_anonymous_allowed(self, allowed, anonuser="anon"):
        if allowed:
            self.anonymous_credential = (anonuser, anonuser)
        else:
            self.anonymous_credential = None

    @property
    def apphome(self):
        return self._apphome

    appdatahome = apphome

    def load_configuration(self, **kw):
        super().load_configuration(**kw)
        # no undo support in tests
        self.global_set_option("undo-enabled", "n")

    def bootstrap_cubes(self):
        try:
            super().bootstrap_cubes()
        except IOError:
            # no cubes
            self.init_cubes(())

    def read_sources_file(self):
        """By default, we run tests with the sqlite DB backend.  One may use its
        own configuration by just creating a 'sources' file in the test
        directory from which tests are launched or by specifying an alternative
        sources file using self.sourcefile.
        """
        if getattr(self, "sourcefile", None):
            raise Exception(
                "sourcefile isn't supported anymore, specify your database "
                "configuration using proper configuration class (e.g. "
                "PostgresApptestConfiguration)"
            )
        try:
            super().read_sources_file()
            raise Exception(
                "test configuration shouldn't provide a sources file, specify your "
                "database configuration using proper configuration class (e.g. "
                "PostgresApptestConfiguration)"
            )
        except ExecutionError:
            pass
        return self.default_sources

    # web config methods needed here for cases when we use this config as a web
    # config

    def default_base_url(self):
        return BASE_URL


class PostgresApptestConfiguration(ApptestConfiguration):
    default_sources = DEFAULT_PSQL_SOURCES



from logilab.common.graph import get_cycles

try:
    (rtype,) = __args__
except ValueError:
    print("USAGE: cubicweb-ctl shell <instance> detect_cycle.py -- <relation type>")
    print()

graph = {}
for fromeid, toeid in rql(f"Any X,Y WHERE X {rtype} Y"):
    graph.setdefault(fromeid, []).append(toeid)

for cycle in get_cycles(graph):
    print("cycle", "->".join(str(n) for n in cycle))



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for selectors mechanism"""

from contextlib import contextmanager
from operator import eq, lt, le, gt

from logilab.common.registry import NoSelectableObject
from logilab.common.decorators import clear_cache
from logilab.common.testlib import TestCase, unittest_main

from cubicweb import Binary
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.entity import EntityAdapter, Adapter
from cubicweb.predicates import (
    is_instance,
    adaptable,
    match_kwargs,
    match_user_groups,
    multi_lines_rset,
    score_entity,
    is_in_state,
    rql_condition,
    relation_possible,
)


class ImplementsTC(CubicWebTC):
    def test_etype_priority(self):
        with self.admin_access.cnx() as cnx:
            f = cnx.create_entity(
                "FakeFile",
                data_name="hop.txt",
                data=Binary(b"hop"),
                data_format="text/plain",
            )
            rset = f.as_rset()
            anyscore = is_instance("Any")(f.__class__, cnx, rset=rset)
            idownscore = adaptable("IDownloadable")(f.__class__, cnx, rset=rset)
            self.assertGreater(idownscore, anyscore, (idownscore, anyscore))
            filescore = is_instance("FakeFile")(f.__class__, cnx, rset=rset)
            self.assertGreater(filescore, idownscore, (filescore, idownscore))

    def test_etype_inheritance_no_yams_inheritance(self):
        cls = self.vreg["etypes"].etype_class("Personne")
        with self.admin_access.cnx() as cnx:
            self.assertFalse(is_instance("Societe").score_class(cls, cnx))

    def test_yams_inheritance(self):
        cls = self.vreg["etypes"].etype_class("Transition")
        with self.admin_access.cnx() as cnx:
            self.assertEqual(is_instance("BaseTransition").score_class(cls, cnx), 3)

    def test_outer_join(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute('Any U,B WHERE B? bookmarked_by U, U login "anon"')
            self.assertEqual(
                is_instance("Bookmark")(None, cnx, rset=rset, row=0, col=1), 0
            )


class WorkflowSelectorTC(CubicWebTC):
    def setUp(self):
        super().setUp()
        # enable debug mode to state/transition validation on the fly
        self.vreg.config.debugmode = True

    def tearDown(self):
        self.vreg.config.debugmode = False
        super().tearDown()

    def setup_database(self):
        with self.admin_access.shell() as shell:
            wf = shell.add_workflow("wf_test", "StateFull", default=True)
            created = wf.add_state("created", initial=True)
            validated = wf.add_state("validated")
            abandoned = wf.add_state("abandoned")
            wf.add_transition("validate", created, validated, ("managers",))
            wf.add_transition(
                "forsake",
                (
                    created,
                    validated,
                ),
                abandoned,
                ("managers",),
            )

    @contextmanager
    def statefull_stuff(self):
        with self.admin_access.cnx() as cnx:
            wf_entity = cnx.create_entity("StateFull", name="")
            rset = wf_entity.as_rset()
            adapter = wf_entity.cw_adapt_to("IWorkflowable")
            cnx.commit()
            self.assertEqual(adapter.state, "created")
            yield cnx, wf_entity, rset, adapter

    def test_is_in_state(self):
        with self.statefull_stuff() as (cnx, wf_entity, rset, adapter):
            for state in ("created", "validated", "abandoned"):
                selector = is_in_state(state)
                self.assertEqual(selector(None, cnx, rset=rset), state == "created")

            adapter.fire_transition("validate")
            cnx.commit()
            wf_entity.cw_clear_all_caches()
            self.assertEqual(adapter.state, "validated")

            clear_cache(rset, "get_entity")

            selector = is_in_state("created")
            self.assertEqual(selector(None, cnx, rset=rset), 0)
            selector = is_in_state("validated")
            self.assertEqual(selector(None, cnx, rset=rset), 1)
            selector = is_in_state("validated", "abandoned")
            self.assertEqual(selector(None, cnx, rset=rset), 1)
            selector = is_in_state("abandoned")
            self.assertEqual(selector(None, cnx, rset=rset), 0)

            adapter.fire_transition("forsake")
            cnx.commit()
            wf_entity.cw_clear_all_caches()
            self.assertEqual(adapter.state, "abandoned")

            clear_cache(rset, "get_entity")

            selector = is_in_state("created")
            self.assertEqual(selector(None, cnx, rset=rset), 0)
            selector = is_in_state("validated")
            self.assertEqual(selector(None, cnx, rset=rset), 0)
            selector = is_in_state("validated", "abandoned")
            self.assertEqual(selector(None, cnx, rset=rset), 1)
            self.assertEqual(adapter.state, "abandoned")
            self.assertEqual(selector(None, cnx, rset=rset), 1)

    def test_is_in_state_unvalid_names(self):
        with self.statefull_stuff() as (cnx, wf_entity, rset, adapter):
            selector = is_in_state("unknown")
            with self.assertRaises(ValueError) as cm:
                selector(None, cnx, rset=rset)
            self.assertEqual(str(cm.exception), "wf_test: unknown state(s): unknown")
            selector = is_in_state("weird", "unknown", "created", "weird")
            with self.assertRaises(ValueError) as cm:
                selector(None, cnx, rset=rset)
            self.assertEqual(
                str(cm.exception), "wf_test: unknown state(s): unknown,weird"
            )


class RelationPossibleTC(CubicWebTC):
    def test_rqlst_1(self):
        with self.admin_access.cnx() as cnx:
            selector = relation_possible("in_group")
            select = self.vreg.parse(cnx, "Any X WHERE X is CWUser").children[0]
            score = selector(
                None,
                cnx,
                rset=1,
                select=select,
                filtered_variable=select.defined_vars["X"],
            )
            self.assertEqual(score, 1)

    def test_rqlst_2(self):
        with self.admin_access.cnx() as cnx:
            selector = relation_possible("in_group")
            select = self.vreg.parse(
                cnx,
                "Any 1, COUNT(X) WHERE X is CWUser, X creation_date XD, "
                "Y creation_date YD, Y is CWGroup "
                "HAVING DAY(XD)=DAY(YD)",
            ).children[0]
            score = selector(
                None,
                cnx,
                rset=1,
                select=select,
                filtered_variable=select.defined_vars["X"],
            )
            self.assertEqual(score, 1)

    def test_ambiguous(self):
        # Ambiguous relations are :
        # (Service, fabrique_par, Personne) and (Produit, fabrique_par, Usine)
        # There used to be a crash here with a bad rdef choice in the strict
        # checking case.
        selector = relation_possible(
            "fabrique_par", role="object", target_etype="Personne", strict=True
        )
        with self.admin_access.cnx() as cnx:
            usine = cnx.create_entity("Usine", lieu="here")
            score = selector(None, cnx, rset=usine.as_rset())
            self.assertEqual(0, score)


class MatchUserGroupsTC(CubicWebTC):
    def test_owners_group(self):
        """tests usage of 'owners' group with match_user_group"""

        class SomeAdapter(Adapter):
            __regid__ = "yo"
            __select__ = match_user_groups("owners")

        self.vreg._loadedmods[__name__] = {}
        self.vreg.register(SomeAdapter)
        SomeAdapter.__registered__(self.vreg["adapters"])
        self.assertIn(SomeAdapter, self.vreg["adapters"]["yo"], self.vreg["adapters"])

        try:
            with self.admin_access.repo_cnx() as cnx:
                self.create_user(cnx, "john")

            # login as a simple user
            john_access = self.new_access("john")
            with john_access.repo_cnx() as cnx:
                # it should not be possible to use SomeAction not owned objects
                rset = cnx.execute('Any G WHERE G is CWGroup, G name "managers"')

                with self.assertRaises(NoSelectableObject):
                    cnx.vreg["adapters"].select("yo", cnx)

                # insert a new card, and check that we can use SomeAction on our object
                cnx.execute('INSERT Card C: C title "zoubidou"')
                cnx.commit()

                rset = cnx.execute('Card C WHERE C title "zoubidou"')

                # shouldn't raise since we have the rights now
                cnx.vreg["adapters"].select("yo", cnx, rset=rset)

            # make sure even managers can't use the action
            with self.admin_access.repo_cnx() as cnx:
                rset = cnx.execute('Card C WHERE C title "zoubidou"')

                with self.assertRaises(NoSelectableObject):
                    cnx.vreg["adapters"].select("yo", cnx, rset=rset)

        finally:
            del self.vreg[SomeAdapter.__registry__][SomeAdapter.__regid__]


class MultiLinesRsetTC(CubicWebTC):
    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            cnx.execute('INSERT CWGroup G: G name "group1"')
            cnx.execute('INSERT CWGroup G: G name "group2"')
            cnx.commit()

    def test_default_op_in_selector(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute("Any G WHERE G is CWGroup")
            expected = len(rset)
            selector = multi_lines_rset(expected)
            self.assertEqual(selector(None, cnx, rset=rset), 1)
            self.assertEqual(selector(None, cnx, None), 0)
            selector = multi_lines_rset(expected + 1)
            self.assertEqual(selector(None, cnx, rset=rset), 0)
            self.assertEqual(selector(None, cnx, None), 0)
            selector = multi_lines_rset(expected - 1)
            self.assertEqual(selector(None, cnx, rset=rset), 0)
            self.assertEqual(selector(None, cnx, None), 0)

    def test_without_rset(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute("Any G WHERE G is CWGroup")
            expected = len(rset)
            selector = multi_lines_rset(expected)
            self.assertEqual(selector(None, cnx, None), 0)
            selector = multi_lines_rset(expected + 1)
            self.assertEqual(selector(None, cnx, None), 0)
            selector = multi_lines_rset(expected - 1)
            self.assertEqual(selector(None, cnx, None), 0)

    def test_with_operators(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute("Any G WHERE G is CWGroup")
            expected = len(rset)

            # Format     'expected', 'operator', 'assert'
            testdata = (
                (expected, eq, 1),
                (expected + 1, eq, 0),
                (expected - 1, eq, 0),
                (expected, le, 1),
                (expected + 1, le, 1),
                (expected - 1, le, 0),
                (expected - 1, gt, 1),
                (expected, gt, 0),
                (expected + 1, gt, 0),
                (expected + 1, lt, 1),
                (expected, lt, 0),
                (expected - 1, lt, 0),
            )

            for expected, operator, assertion in testdata:
                selector = multi_lines_rset(expected, operator)
                with self.subTest(expected=expected, operator=operator):
                    self.assertEqual(selector(None, cnx, rset=rset), assertion)


class MatchKwargsTC(TestCase):
    def test_match_kwargs_default(self):
        selector = match_kwargs({"a", "b"})
        self.assertEqual(selector(None, None, a=1, b=2), 2)
        self.assertEqual(selector(None, None, a=1), 0)
        self.assertEqual(selector(None, None, c=1), 0)
        self.assertEqual(selector(None, None, a=1, c=1), 0)

    def test_match_kwargs_any(self):
        selector = match_kwargs({"a", "b"}, mode="any")
        self.assertEqual(selector(None, None, a=1, b=2), 2)
        self.assertEqual(selector(None, None, a=1), 1)
        self.assertEqual(selector(None, None, c=1), 0)
        self.assertEqual(selector(None, None, a=1, c=1), 1)


class ScoreEntityTC(CubicWebTC):
    def test_intscore_entity_selector(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute("Any E WHERE E eid 1")
            selector = score_entity(lambda x: None)
            self.assertEqual(selector(None, cnx, rset=rset), 0)
            selector = score_entity(lambda x: "something")
            self.assertEqual(selector(None, cnx, rset=rset), 1)
            selector = score_entity(lambda x: object)
            self.assertEqual(selector(None, cnx, rset=rset), 1)
            rset = cnx.execute("Any G LIMIT 2 WHERE G is CWGroup")
            selector = score_entity(lambda x: 10)
            self.assertEqual(selector(None, cnx, rset=rset), 20)
            selector = score_entity(lambda x: 10, mode="any")
            self.assertEqual(selector(None, cnx, rset=rset), 10)

    def test_rql_condition_entity(self):
        with self.admin_access.cnx() as cnx:
            selector = rql_condition("X identity U")
            rset = cnx.user.as_rset()
            self.assertEqual(selector(None, cnx, rset=rset), 1)
            self.assertEqual(selector(None, cnx, entity=cnx.user), 1)
            self.assertEqual(selector(None, cnx), 0)

    def test_rql_condition_user(self):
        with self.admin_access.cnx() as cnx:
            selector = rql_condition('U login "admin"', user_condition=True)
            self.assertEqual(selector(None, cnx), 1)
            selector = rql_condition('U login "toto"', user_condition=True)
            self.assertEqual(selector(None, cnx), 0)


class AdaptablePredicateTC(CubicWebTC):
    def test_multiple_entity_types_rset(self):
        class CWUserIWhatever(EntityAdapter):
            __regid__ = "IWhatever"
            __select__ = is_instance("CWUser")

        class CWGroupIWhatever(EntityAdapter):
            __regid__ = "IWhatever"
            __select__ = is_instance("CWGroup")

        with self.temporary_appobjects(CWUserIWhatever, CWGroupIWhatever):
            with self.admin_access.cnx() as cnx:
                selector = adaptable("IWhatever")
                rset = cnx.execute("Any X WHERE X is IN(CWGroup, CWUser)")
                self.assertTrue(selector(None, cnx, rset=rset))


if __name__ == "__main__":
    unittest_main()
